
#include "stdafx.h"
#include "XmlFileManager.h"

#include "AppLogic.h"
#include "uuidStats.h" // CLSIDs

#include "HtmlParser.h"

////////////////////////////////////////////////////////////////////////////////////////
//
//
// CAppLogic class
//
//
/////////////////////////////////////////////////////////////////////////////////////////
// Stephane Rodriguez, 1st july 2002
//
//




CAppLogic::CAppLogic() : m_cRef(0)
{
	init();
}


CAppLogic::CAppLogic(const CAppLogic &obj) : m_cRef(0)
{
	init();
}


CAppLogic::~CAppLogic()
{
	clean();
}


CAppLogic& CAppLogic::operator=(const CAppLogic &obj)
{
	// prohibited use
	return *this;
}


long CAppLogic::m_nLastError;



void CAppLogic::init()
{
	m_pRoot = NULL;
	setParent(NULL);
	
	initGeneralStats();

	XmlhashTableInit(&m_hDuplicatedValues);
}



HRESULT CAppLogic::QueryInterface(const IID& iid, void** ppv) // IXmlStats interface support
{ 	
	if (iid == IID_IUnknown)
	{
		*ppv = static_cast<IXmlStats*>(this) ;
	} 
	else if (iid == IID_IXmlStats)
	{
		*ppv = static_cast<IXmlStats*>(this) ;
	}
	else
	{  	   
		*ppv = NULL ;
		return E_NOINTERFACE;
	}
	reinterpret_cast<IUnknown*>(*ppv)->AddRef() ; 
	return S_OK ;
}

ULONG CAppLogic::AddRef()
{
	return InterlockedIncrement(&m_cRef) ;
}

ULONG CAppLogic::Release() 
{
	if (InterlockedDecrement(&m_cRef) == 0)
	{
		delete this ;
		return 0 ;
	}
	return m_cRef ;
}



void  CAppLogic::setParent(void *pParent)
{
	m_cpParent = pParent;
}

void* CAppLogic::getParent()
{
	return m_cpParent;
}


// used by callbacks. Declared here because of C constraints.
char g_szContent[16384];
long g_szContentLength;
CElement *g_cpCurrentElement;
BOOL g_bContent;


BOOL CAppLogic::process(char *szFilename)
{
	BOOL bError = TRUE;

	clean();

	if (szFilename==NULL || strlen(szFilename)==0)
	{
		setLastError(SAXML_ERROR_BADURL);
		trace2(TRC_ERROR,"CAppLogic::process",SAXML_ERRORSTRING_BADURL);
		return FALSE;
	}

	m_szFilename = szFilename; // store filename somewhere (filename can be of the form http://...)

	m_bExternalForcedStop = FALSE;

	CXmlFileManager *pResourceManager = CXmlFileManager::chooseResourceManager(szFilename);

	if (!pResourceManager)
	{
		setLastError( SAXML_ERROR_OUTOFMEMORY );
		trace2(TRC_ERROR,"CAppLogic::process, pResourceManager",SAXML_ERRORSTRING_OUTOFMEMORY);
		return FALSE;
	}

	if (!m_pRoot) m_pRoot = new CElement (this, NULL, "");
	if (!m_pRoot) 
	{
		setLastError( SAXML_ERROR_OUTOFMEMORY );
		trace2(TRC_ERROR,"CAppLogic::process, m_pRoot",SAXML_ERRORSTRING_OUTOFMEMORY);
		delete pResourceManager;
		return FALSE;
	}

	// transmit caller to let ResourceManager post(receive) messages
	pResourceManager->setParent( getParent() );

	if (!pResourceManager->Open(szFilename))
	{
		setLastError( pResourceManager->getLastError() );
		trace2(TRC_ERROR,"CAppLogic::process, error in Resource manager\n","");
		delete pResourceManager;
		return FALSE;
	}

	// read the XML file, and parse the elements/attributes
	//
	
	char szBuf[SAXML_BUFSIZE];
	m_parser = XML_ParserCreate(NULL); // create an expat parser instance
	BOOL bStop;
	int len;


	XML_SetUserData(m_parser, m_pRoot); // set user_data that will be passed in all callback functions

	XML_SetElementHandler(m_parser, cbk_startElement, cbk_endElement);
	XML_SetCharacterDataHandler(m_parser, cbk_contentElement);
	XML_SetCommentHandler(m_parser, cbk_commenthandler);
	XML_SetCdataSectionHandler(m_parser, cbk_cdataSection, cbk_endcdataSection);
	XML_SetProcessingInstructionHandler(m_parser, cbk_pi);

	g_szContentLength = 0;
	g_cpCurrentElement = NULL;

	initGeneralStats();
	initDepthStats();
	initPatternStats();
	initIndentationStats();
	initContentStats();
	initStructStats();

	do 
	{
		if (!pResourceManager->Read(szBuf,sizeof(szBuf),len))
		{
			len = 0;
			setLastError( pResourceManager->getLastError() );
			trace2(TRC_ERROR,"CAppLogic::process","");
		}

		// count nblines in this buffer
		if (len>0)
		{
			addStatsStreamNbLines( countNbLines(szBuf,len) );

			// also count multiple spaces
			char *s = new char[len+1];
			strncpy(s,szBuf,len);
			s[len]='\0'; // add EOL to make sure that strlen(s) works fine

			HTMLParser html;
			UINT nStart=0; // index where to start html parsing
			UINT nFirst; // where there is a '<' sign
			UINT nEnd; // where there is a '>' sign
			UINT nNbMultipleSpaces = 0;

			while ( html.GetNextTag(s,nStart,/*out*/nFirst,/*out*/nEnd,/*out*/nNbMultipleSpaces) )
			{
				m_nNbMultipleSpaces += nNbMultipleSpaces;
				nNbMultipleSpaces = 0;

				// iterate
				nStart = nEnd+1;
			}

			m_nNbMultipleSpaces += nNbMultipleSpaces;

			delete [] s;
		}

		bStop = ( len==0 || pResourceManager->shouldSkipNow() );

		if (!bStop && !XML_Parse(m_parser, szBuf, len, bStop)) 
		{
			char szError[MAX_PATH];
			sprintf(szError,"%s at line %d\n",
				XML_ErrorString(XML_GetErrorCode(m_parser)),
				XML_GetCurrentLineNumber(m_parser));
			trace (TRC_ERROR,szError);

			setLastError( SAXML_ERROR_WHILEPARSING );
			trace2(TRC_ERROR,"CAppLogic::process",SAXML_ERRORSTRING_WHILEPARSING);

			setLastError( 5000 + XML_GetCurrentLineNumber(m_parser) ); // 5000 + line where 

			bError = FALSE;

			break;
		}

		if (!bStop)
			addStatsStreamBufferSize( len );

	} while (!bStop && !m_bExternalForcedStop);

	XML_ParserFree(m_parser);


	setStatsStreamIsFull( (len==0) ? TRUE : FALSE );

	if (bError)
	{
		evalStructStatsMinMaxMean( m_pRoot ); // produces many things, and a value for m_nPatternElementMaxOccurences

		evalStructStatsMinMaxMeanAttribsPerElement( m_pRoot );

		long nBiasThreshold = m_nPatternElementMaxOccurences*80/100; // 20% bias
		evalMostSeenPattern( m_pRoot, nBiasThreshold );
		m_nbPatterns=0;
		findMostSeenPattern( m_pRoot, m_nbPatterns);

		evalDatatypes( m_pRoot );

		evalContentStatsMinMaxMeanDuplicates( m_pRoot );
	}

	delete pResourceManager; // close resource

	return bError;

}


void CAppLogic::forceStop() // command to stop processing inside process()
{
	m_bExternalForcedStop = TRUE; // allows to stop
}


void CAppLogic::initGeneralStats()
{
	// init general stats
	m_nStreamLength = m_nStreamNbLines = 0;
	m_nNbElements = m_nNbAttributes = m_nNbComments = m_nNbCDATASections = m_nNbPI = 0;
	m_nCommentSize = m_nCDATASectionsSize = 0;
	setStatsStreamIsFull(FALSE);
	setCDATASectionMode(FALSE);
	m_nUseNamespaces = 0;
}

void CAppLogic::setStatsStreamIsFull(BOOL bFull)
{
	m_bStreamFull = bFull;
}

void CAppLogic::addStatsStreamBufferSize(long nBufferSize)
{
	m_nStreamLength += nBufferSize;
}

void CAppLogic::addStatsStreamNbLines(long nIncr)
{
	m_nStreamNbLines += nIncr;
}

void CAppLogic::addStatsNbElements(long nIncr)
{
	m_nNbElements += nIncr;
}

void CAppLogic::addStatsNbAttributes(long nIncr)
{
	m_nNbAttributes += nIncr;
}

void CAppLogic::addStatsNbComments(long nIncr, long nSize)
{
	m_nNbComments += nIncr;
	m_nCommentSize += nSize;
}

void CAppLogic::addStatsCDATASections(long nIncr, long nSize)
{
	m_nNbCDATASections += nIncr;
	m_nCDATASectionsSize += nSize;
}

void CAppLogic::setCDATASectionMode(BOOL bActivate)
{
	m_bCDATASectionMode = bActivate;
}

BOOL CAppLogic::isCDATASectionMode()
{
	return m_bCDATASectionMode;
}

void CAppLogic::addStatsNbPI(long nIncr)
{
	m_nNbPI += nIncr;
}


APICALL char *CAppLogic::getStatsStreamFilename()
{
	return m_szFilename.GetData();
}

APICALL long CAppLogic::getStatsStreamLength()
{
	return m_nStreamLength;
}

APICALL BOOL CAppLogic::getStatsStreamIsFull()
{
	return m_bStreamFull;
}

APICALL long CAppLogic::getStatsNbLines()
{
	return m_nStreamNbLines;
}

APICALL long CAppLogic::getStatsNbElements()
{
	return m_nNbElements;
}

APICALL long CAppLogic::getStatsNbAttributes()
{
	return m_nNbAttributes;
}

APICALL long CAppLogic::getStatsNbComments()
{
	return m_nNbComments;
}

APICALL long CAppLogic::getStatsCommentsSize()
{
	return m_nCommentSize;
}

APICALL long CAppLogic::getStatsCDATASections()
{
	return m_nNbCDATASections;
}

APICALL long CAppLogic::getStatsCDATASectionsSize()
{
	return m_nCDATASectionsSize;
}

APICALL long CAppLogic::getStatsNbPI()
{
	return m_nNbPI;
}




void CAppLogic::initDepthStats()
{
	setCurDepth(0);
	m_arrDepthOccurences.RemoveAll();
	m_arrDepthOccurences.SetSize(128); // 128 default depth to avoid realloc during parsing, will stretch dynamically 
}

void CAppLogic::setCurDepth(long n)
{
	m_nCurDepth = 0;
}

void CAppLogic::addCurDepth(long nIncr)
{
	// add an occurence
	if (nIncr>0 && m_nCurDepth>-1)
	{
		if (m_nCurDepth < getDepthMax())
			m_arrDepthOccurences.SetAt( m_nCurDepth, m_arrDepthOccurences.GetAt(m_nCurDepth) + 1);
		else
			m_arrDepthOccurences.SetAtGrow(m_nCurDepth, 1); // initialize depth occurence
	}

	m_nCurDepth += nIncr;
}

long CAppLogic::getDepthMax() // 1-based
{
	return m_arrDepthOccurences.GetUpperBound();
}

double CAppLogic::getDepthMean() // 1-based
{
	// calculation follows
	double m = 0.0;

	long nSize = getDepthMax(),i;
	long nTotal = 0;
	long n;

	for (i=0;i<nSize;i++)
	{
		n = m_arrDepthOccurences.GetAt(i);
		nTotal += n;
		m += n*i;
	}

	if (nTotal==0)
		return m; // in fact, that should be N/A instead of 0

	m /= nTotal;

	return m;
}

#define dabs(a,b) ((a>b)?(a-b):(b-a))

double CAppLogic::getDepthStdDev()
{
	// calculation follows
	double s = 0.0;

	long nSize = getDepthMax(),i;
	long nTotal = 0;
	long n;

	double m = getDepthMean();

	for (i=0;i<nSize;i++)
	{
		n = m_arrDepthOccurences.GetAt(i);
		nTotal += n;
		s += n * dabs( i , m );
	}

	if (nTotal==0)
		return s; // in fact, that should be N/A instead of 0

	s /= nTotal;

	return s;
}

long CAppLogic::getDepthOccurences(long nDepth) // depth is 1-based
{
	if (nDepth>=1 && nDepth<=getDepthMax())
		return m_arrDepthOccurences.GetAt( nDepth-1 );
	else
		return 0;
}

// sends back the collection of all element nodes. (collection is built by copying ptrs to nodes. DO NOT destroy these nodes yourself!)
APICALL void CAppLogic::getDepthElementCollection(long nDepth, /*out*/XmlArrayNode &elementCollection)
{
	if (!m_pRoot)
		return; // there must be something, stupid!

	long nCurDepth = 0;
	long nTargetDepth = nDepth;

	getDepthElementCollection( m_pRoot, nTargetDepth, nCurDepth, elementCollection);
}


void CAppLogic::getDepthElementCollection(IXmlNode *p, long nTargetDepth, long nCurDepth, /*out*/XmlArrayNode &elementCollection)
{

	if (!p)
		return;

	CElement *pElement = (CElement *)p;

	// add contribution of this element

	if (nCurDepth == nTargetDepth)
		elementCollection.Add( p );

	// then process all direct element children
	//
	int nChildElementsSize = pElement->getChildrenSize();
	for (int k=0; k<nChildElementsSize; k++)
	{
		CElement *pChildElement = pElement->getChildrenElement(k);
		if (pChildElement)
			getDepthElementCollection(pChildElement, nTargetDepth, nCurDepth+1, elementCollection);
	}


}



void CAppLogic::initStructStats()
{
	m_arrStructNameOccurences.RemoveAll();
	m_arrStructNameSizes.RemoveAll();

	m_nPatternElementMaxOccurences = 0; // calculated in struct stats, but used in pattern stats

	m_nMinNbAttribsPerElement = 0x7FFFFFFF;
	m_nMaxNbAttribsPerElement = m_nTotalNbAttribsPerElement = 0;
	m_ndStdDevNbAttribsPerElement = 0.0;

	m_nTotalAttribsDisorder = 0;

}

void CAppLogic::setStructNameOccurences(long n, long nOccurences)
{
	if (n<m_arrStructNameOccurences.GetSize())
		m_arrStructNameOccurences.SetAt(n, nOccurences+m_arrStructNameOccurences.GetAt(n) ); //add
	else
		m_arrStructNameOccurences.SetAtGrow(n, nOccurences); //set
}

void CAppLogic::setStructNameSizes(long n, long nSize) // total size in bytes within struct of "element with size n"
{
	if (n<m_arrStructNameSizes.GetSize())
		m_arrStructNameSizes.SetAt(n, nSize+m_arrStructNameSizes.GetAt(n) ); //add
	else
		m_arrStructNameSizes.SetAtGrow(n, nSize); //set
}


void CAppLogic::evalStructStatsMinMaxMean(IXmlNode *p)
{
	if (!p)
		return;

	CElement *pElement = (CElement *)p;

	// add contribution of this element

	long n = strlen( p->getName() );
	long occ = p->getAmountOccurences();

	m_nPatternElementMaxOccurences = max(m_nPatternElementMaxOccurences, occ); // forward code for pattern stats
	p->setMostSeen(FALSE); // forward code for pattern stats

	setStructNameOccurences(n, occ);
	setStructNameSizes(n, ((p->getHasContent()==SAXML_CONTENT_YES)||(pElement->getChildrenSize()>0)) ? n*occ*2 : n*occ); // size is doubled if we are working with a element with content because it has a /tag

	// process all attributes
	//
	long nAttributes = pElement->getAttributes().GetSize();
	for (long j=0; j<nAttributes; j++)
	{
		CAttribute *pAttribute = (CAttribute *) pElement->getAttributes().GetAt(j);
		if (pAttribute)
		{
			n = strlen( pAttribute->getName() );
			long occ = pAttribute->getAmountOccurences();

			m_nPatternElementMaxOccurences = max(m_nPatternElementMaxOccurences, occ); // forward code for pattern stats
			pAttribute->setMostSeen(FALSE); // forward code for pattern stats

			setStructNameOccurences(n, occ);
			setStructNameSizes(n, n*occ);
		}
	} // end process all attributes


	// then process all direct element children
	//
	int nChildElementsSize = pElement->getChildrenSize();
	for (int k=0; k<nChildElementsSize; k++)
	{
		CElement *pChildElement = pElement->getChildrenElement(k);
		if (pChildElement)
			evalStructStatsMinMaxMean(pChildElement);
	}

}

void CAppLogic::evalStructStatsIfNodesSizesOptimized(IXmlNode *p, long &i, long &nTotal)
{
	if (!p)
		return;

	CElement *pElement = (CElement *)p;

	// add contribution of this element

	long n = (i<26) ? 1 : (i-26)/36+2;
	long occ = p->getAmountOccurences();

	i++;

	nTotal +=  ((p->getHasContent()==SAXML_CONTENT_YES)||(pElement->getChildrenSize()>0)) ? n*occ*2 : n*occ; // size is doubled if we are working with a element with content because it has a /tag

	// process all attributes
	//
	long nAttributes = pElement->getAttributes().GetSize();
	for (long j=0; j<nAttributes; j++)
	{
		CAttribute *pAttribute = (CAttribute *) pElement->getAttributes().GetAt(j);
		if (pAttribute)
		{
			n = (i<26) ? 1 : (i-26)/36+2;
			long occ = pAttribute->getAmountOccurences();

			i++;

			nTotal += n*occ;
		}
	} // end process all attributes


	// then process all direct element children
	//
	int nChildElementsSize = pElement->getChildrenSize();
	for (int k=0; k<nChildElementsSize; k++)
	{
		CElement *pChildElement = pElement->getChildrenElement(k);
		if (pChildElement)
			evalStructStatsIfNodesSizesOptimized(pChildElement,i,nTotal);
	}
}

long CAppLogic::getStructStatsTotalOccurences()
{
	long nSize = m_arrStructNameOccurences.GetUpperBound();
	if (nSize==0)
		return 0;

	long nTotal = 0;
	for (long i=0;i<nSize;i++)
		nTotal += m_arrStructNameOccurences.GetAt(i);

	return nTotal;
}

long CAppLogic::getStructStatsTotalSize()
{
	long nSize = m_arrStructNameSizes.GetUpperBound();
	if (nSize==0)
		return 0;

	long nTotal = 0;
	for (long i=0;i<nSize;i++)
		nTotal += m_arrStructNameSizes.GetAt(i);

	return nTotal;

}

long CAppLogic::getStructStatsMin()
{
	long nSize = m_arrStructNameOccurences.GetUpperBound();
	if (nSize==0)
		return 0;

	long i = 0;
	while ( i<nSize && m_arrStructNameOccurences.GetAt(i)==0 )
		i++;

	if (i>=nSize)
		return 0; // not found

	return i;
}

long CAppLogic::getStructStatsMax()
{
	long nSize = m_arrStructNameOccurences.GetUpperBound();
	if (nSize==0)
		return 0;

	return nSize-1;
}

double CAppLogic::getStructStatsMean()
{
	long nSize = m_arrStructNameOccurences.GetUpperBound();
	if (nSize==0)
		return 0.0;

	// calculation follows
	double m = 0.0;

	long i,n;
	long nTotal = 0;

	for (i=0;i<nSize;i++)
	{
		n = m_arrStructNameOccurences.GetAt(i);
		nTotal += n;
		m += n*i;
	}

	if (nTotal==0)
		return m; // in fact, that should be N/A instead of 0

	m /= nTotal;

	return m;
}

double CAppLogic::getStructStatsStdDev()
{
	long nSize = m_arrStructNameOccurences.GetUpperBound();
	if (nSize==0)
		return 0.0;

	// calculation follows
	double s = 0.0;

	long i,n;
	long nTotal = 0;

	double m = getStructStatsMean();

	for (i=0;i<nSize;i++)
	{
		n = m_arrStructNameOccurences.GetAt(i);
		nTotal += n;
		s += n * dabs( i , m );
	}

	if (nTotal==0)
		return s; // in fact, that should be N/A instead of 0

	s /= nTotal;

	return s;
}

long CAppLogic::getStructStatsNameOccurences(long i) // i is the length in bytes of the requested elem+attrib info
{
	long nSize = m_arrStructNameOccurences.GetUpperBound();
	if (i>=nSize)
		return 0;
	else
		return m_arrStructNameOccurences.GetAt(i);
}

long CAppLogic::getStructStatsNameSizes(long i) // i is the length in bytes of the requested elem+attrib info
{
	long nSize = m_arrStructNameSizes.GetUpperBound();
	if (i>=nSize)
		return 0;
	else
		return m_arrStructNameSizes.GetAt(i);
}


APICALL void CAppLogic::getStructStatsNamesCollection(/*size of names in bytes*/long i, /*out*/XmlArrayString &nameCollection) // all elements+attribs which name is of length i bytes
{
	getStructStatsNamesCollection(m_pRoot, i, nameCollection);
}

void CAppLogic::getStructStatsNamesCollection(IXmlNode *p, /*size of names in bytes*/long i, /*out*/XmlArrayString &nameCollection) // all elements+attribs which name is of length i bytes
{
	if (!p)
		return;

	CElement *pElement = (CElement *)p;

	// add contribution of this element
	if (m_pRoot!=pElement)
	{
		long n = strlen( p->getName() );
		if (n==i)
			nameCollection.Add( new String( p->getName() ) );
	}

	// process all attributes
	//
	long nAttributes = pElement->getAttributes().GetSize();
	for (long j=0; j<nAttributes; j++)
	{
		CAttribute *pAttribute = (CAttribute *) pElement->getAttributes().GetAt(j);
		if (pAttribute)
		{
			long n = strlen( pAttribute->getName() );
			if (n==i)
				nameCollection.Add( new String( pAttribute->getName() ) );
		}
	} // end process all attributes


	// then process all direct element children
	//
	int nChildElementsSize = pElement->getChildrenSize();
	for (int k=0; k<nChildElementsSize; k++)
	{
		CElement *pChildElement = pElement->getChildrenElement(k);
		if (pChildElement)
			getStructStatsNamesCollection(pChildElement,i,nameCollection);
	}

}


APICALL long CAppLogic::getStructStatsMinNbAttributsPerElement()
{
	return m_nMinNbAttribsPerElement;
}
APICALL long CAppLogic::getStructStatsMaxNbAttributsPerElement()
{
	return m_nMaxNbAttribsPerElement;
}
APICALL double CAppLogic::getStructStatsMeanNbAttributsPerElement()
{
	if (!m_pRoot)
		return 0.0;

	long nbElements = getStatsNbElements(); //nb elements in structure, not all occurences in the overall stream, -1 because __root__ shouldn't appear

	return nbElements==0 ? 0.0 : double(m_nTotalNbAttribsPerElement)/nbElements;
}
APICALL double CAppLogic::getStructStatsStdDevNbAttributsPerElement()
{
	if (!m_pRoot)
		return 0.0;

	long nbElements = getStatsNbElements(); //nb elements in structure, not all occurences in the overall stream, -1 because __root__ shouldn't appear

	evalStructStatsStdDevAttribsPerElement( m_pRoot, getStructStatsMeanNbAttributsPerElement() );

	return nbElements==0 ? 0.0 : double(m_ndStdDevNbAttribsPerElement)/nbElements;
}

void CAppLogic::evalStructStatsMinMaxMeanAttribsPerElement(IXmlNode *p)
{

	if (!p)
		return;

	CElement *pElement = (CElement *)p;

	// add contribution of this element

	long nAttributes = pElement->getAttributes().GetSize();
	m_nMinNbAttribsPerElement = min(m_nMinNbAttribsPerElement, nAttributes);
	m_nMaxNbAttribsPerElement = max(m_nMaxNbAttribsPerElement, nAttributes);
	m_nTotalNbAttribsPerElement += nAttributes * pElement->getAmountOccurences();


	// then process all direct element children
	//
	int nChildElementsSize = pElement->getChildrenSize();
	for (int k=0; k<nChildElementsSize; k++)
	{
		CElement *pChildElement = pElement->getChildrenElement(k);
		if (pChildElement)
			evalStructStatsMinMaxMeanAttribsPerElement(pChildElement);
	}

}

void CAppLogic::evalStructStatsStdDevAttribsPerElement(IXmlNode *p, double ndMean)
{

	if (!p)
		return;

	CElement *pElement = (CElement *)p;

	// add contribution of this element
	m_ndStdDevNbAttribsPerElement += dabs(pElement->getAttributes().GetSize(),ndMean) * pElement->getAmountOccurences();

	// then process all direct element children
	//
	int nChildElementsSize = pElement->getChildrenSize();
	for (int k=0; k<nChildElementsSize; k++)
	{
		CElement *pChildElement = pElement->getChildrenElement(k);
		if (pChildElement)
			evalStructStatsStdDevAttribsPerElement(pChildElement,ndMean);
	}

}

void CAppLogic::addStructStatsTotalAttribsDisorder(long nIncr)
{
	m_nTotalAttribsDisorder += nIncr;
}

APICALL double CAppLogic::getStructStatsAttribsDisorderRatio()
{
	long nbAttribs = getStatsNbAttributes();

	return nbAttribs==0 ? 0.0 : double(m_nTotalAttribsDisorder)/nbAttribs;
}


void CAppLogic::addStatsUseNamespaces(const char *szName)
{
	String st = (char*) szName;
	long nIndex = st.Find(0,':');
	if (nIndex<=0)
		return;

	do
	{
		nIndex--;
	}
	while (nIndex>0 && st[nIndex]==' ');

	String sN;
	sN = st.Left(nIndex+1);

	if (sN.CompareNoCase(String("xmlns"))) // namespace declaration, not useful by itself
		return; 

	long nSize = m_arrNamespaces.GetSize();
	BOOL bFound = FALSE;
	for (long i=0;i<nSize && !bFound;i++) // simple cloneage
	{
		String *s = m_arrNamespaces.GetAt(i);
		if (s)
			bFound = (*s==sN);
	}
	if (!bFound) // add it then
	{
		m_arrNamespaces.Add( new String( sN.GetData() ) );
	}

	m_nUseNamespaces++;
}

APICALL long CAppLogic::getStatsNbNamespaces()
{
	return m_arrNamespaces.GetSize();
}

APICALL double CAppLogic::getNamespacesUseRatio() // %
{
	long nb = getStatsNbElements() + getStatsNbAttributes();

	return nb==0 ? 0.0 : double(m_nUseNamespaces)/nb;
}

APICALL void CAppLogic::getNamespaces(XmlArrayString &namecoll)
{
	long nSize = m_arrNamespaces.GetSize();
	for (long i=0;i<nSize;i++) // simple cloneage
	{
		String *s = m_arrNamespaces.GetAt(i);
		if (s)
			namecoll.Add( new String( s->GetData() ) );
	}
}




void CAppLogic::initPatternStats()
{
	m_nPatternElementMaxOccurences = m_nbPatterns = 0;
}

APICALL long CAppLogic::getNbPatterns()
{
	return m_nbPatterns; // filled by findMostSeenPattern()
}

APICALL long CAppLogic::getPatternOccurences() // how many times the pattern appears in the stream
{
	return m_nPatternElementMaxOccurences; // maintained by evalStructStatsMinMaxMean()
}

APICALL long CAppLogic::getPatternHeight() // height in nb of lines of the pattern
{
	long nHeight = 0, nSize;
	getRawPatternSize( m_pRoot, /*normal*/FALSE, nSize, nHeight);
	return nHeight;
}

APICALL long CAppLogic::getPatternSize()
{
	long nSize = 0, nHeight;
	getRawPatternSize( m_pRoot, /*normal*/FALSE, nSize, nHeight);
	return nSize;
}

APICALL long CAppLogic::getPatternFlatSize()
{
	long nSize = 0, nHeight;
	getRawPatternSize( m_pRoot, /*flatten*/TRUE, nSize, nHeight);
	return nSize;
}



void CAppLogic::evalMostSeenPattern(IXmlNode *p, long &nThreshold)
{

	if (!p)
		return;

	CElement *pElement = (CElement *)p;

	// add contribution of this element

	if ( p->getAmountOccurences() > nThreshold )
		p->setMostSeen(TRUE);

	// process all attributes
	//
	long nAttributes = pElement->getAttributes().GetSize();
	for (long j=0; j<nAttributes; j++)
	{
		CAttribute *pAttribute = (CAttribute *) pElement->getAttributes().GetAt(j);
		if (pAttribute)
		{

			if ( pAttribute->getAmountOccurences() > nThreshold )
				pAttribute->setMostSeen(TRUE);
		}
	} // end process all attributes


	// then process all direct element children
	//
	int nChildElementsSize = pElement->getChildrenSize();
	for (int k=0; k<nChildElementsSize; k++)
	{
		CElement *pChildElement = pElement->getChildrenElement(k);
		if (pChildElement)
			evalMostSeenPattern(pChildElement,nThreshold);
	}


}



void CAppLogic::findMostSeenPattern(IXmlNode *p, long &nNbPatterns)
{

	if (!p)
		return;

	CElement *pElement = (CElement *)p;

	// add contribution of this element

	BOOL bIsMostSeen = p->isMostSeen();
	if ( bIsMostSeen )
	{
		CElement *pParent = pElement->getParent();
		if (pParent && !pParent->isMostSeen())
			nNbPatterns++; // declare a new pattern here

		if (nNbPatterns>1) // more than one pattern ?  just scratch the other patterns. We just want one !
		{
			bIsMostSeen = FALSE;
			p->setMostSeen(bIsMostSeen);
		}
	}

	// process all attributes
	//
	long nAttributes = pElement->getAttributes().GetSize();
	for (long j=0; j<nAttributes; j++)
	{
		CAttribute *pAttribute = (CAttribute *) pElement->getAttributes().GetAt(j);
		if (pAttribute)
		{
		}
	} // end process all attributes


	// then process all direct element children
	//
	int nChildElementsSize = pElement->getChildrenSize();
	for (int k=0; k<nChildElementsSize; k++)
	{
		CElement *pChildElement = pElement->getChildrenElement(k);
		if (pChildElement)
			findMostSeenPattern(pChildElement,nNbPatterns);
	}


}


void CAppLogic::showMostSeenPattern(IXmlNode *p, long nCurDepth)
{
	if (!p)
		return;

	char szTemp[MAX_PATH];
	char szT[MAX_PATH]="";

	CElement *pElement = (CElement *)p;

	// add contribution of this element
	BOOL bIsMostSeen = p->isMostSeen();
	if ( bIsMostSeen )
	{
		szTemp[0]='\0';
		for (long i=0;i<nCurDepth;i++) // indent <...> tag
			strcat(szTemp," ");
		strcat(szT,szTemp);

		sprintf(szTemp,"<%s", p->getName());
		strcat(szT,szTemp);
		if (p->getDatatype()!=SAXML_DATATYPE_STRING)
		{
			strcat(szT,"(=");
			strcat(szT,p->getDatatypeAsString());
			strcat(szT,")");
		}
	}

	// process all attributes
	//
	long nAttributes = pElement->getAttributes().GetSize();
	for (long j=0; j<nAttributes; j++)
	{
		CAttribute *pAttribute = (CAttribute *) pElement->getAttributes().GetAt(j);
		if (pAttribute && bIsMostSeen)
		{
			sprintf(szTemp," %s", pAttribute->getName());
			strcat(szT,szTemp);
			if (pAttribute->getDatatype()!=SAXML_DATATYPE_STRING)
			{
				strcat(szT,"(=");
				strcat(szT,pAttribute->getDatatypeAsString());
				strcat(szT,")");
			}

		}
	} // end process all attributes

	if ( bIsMostSeen )
	{
		sprintf(szTemp,">\n"); // close tag
		strcat(szT,szTemp);
		trace(TRC_INFO,szT);
	}

	// then process all direct element children
	//
	int nChildElementsSize = pElement->getChildrenSize();
	for (int k=0; k<nChildElementsSize; k++)
	{
		CElement *pChildElement = pElement->getChildrenElement(k);
		if (pChildElement)
			showMostSeenPattern(pChildElement,nCurDepth+1);
	}

}


void CAppLogic::getRawPatternSize(IXmlNode *p, /*in*/BOOL bFlatten, /*out*/long &nSize, /*out*/long &nHeight)
{

	if (!p)
		return;

	CElement *pElement = (CElement *)p;

	// add contribution of this element

	BOOL bIsMostSeen = p->isMostSeen();
	if ( bIsMostSeen )
	{
		long n = strlen( p->getName() );

// evaluating pattern size
		// n = sizeof ("elem") (= 4 here)
		//<elem name="".../> cnt=+n ; conditions = no content AND no children
		//
		//<elem ...>ryytt    cnt=+2*n ; conditions = content AND no children
		//</elem>
		//
		//<elem ...>rttt</elem> cnt=+2*n ; conditions = content AND no children
		//
		//<elem>				cnt=+2*n ; conditions = children
		//	<ze>...
		//</elem>

		nSize += ((p->getHasContent()==SAXML_CONTENT_YES)||(pElement->getChildrenSize()>0)) ? (bFlatten?n:n*2) : n; // size is doubled if we are working with a element with content because it has a /tag

// evaluating pattern height
		//<elem name="".../> cnt=+1 ; conditions = no content AND no children
		//
		//<elem ...>ryytt    cnt=+2 ; conditions = content AND no children
		//</elem>
		//
		//<elem ...>rttt</elem> cnt=+1 ; conditions = content AND no children
		//
		//<elem>				cnt=+2 ; conditions = children
		//	<ze>...
		//</elem>

		// NB : a more accurate pattern height is obtained by taking into account existing EOLs, which is unstable in practice because it may vary between 2 pattern occurences.

		nHeight += (pElement->getChildrenSize()>0) ? 2 : 1; // height of this element is assumed to be one line (ie optimized), unless this element has children
		
	}

	// process all attributes
	//
	long nAttributes = pElement->getAttributes().GetSize();
	for (long j=0; j<nAttributes; j++)
	{
		CAttribute *pAttribute = (CAttribute *) pElement->getAttributes().GetAt(j);
		if (pAttribute)
		{
			long n = strlen( pAttribute->getName() );
			nSize += n;
		}
	} // end process all attributes


	// then process all direct element children
	//
	int nChildElementsSize = pElement->getChildrenSize();
	for (int k=0; k<nChildElementsSize; k++)
	{
		CElement *pChildElement = pElement->getChildrenElement(k);
		if (pChildElement)
			getRawPatternSize(pChildElement,bFlatten,nSize,nHeight);
	}

}





void CAppLogic::initIndentationStats()
{
	m_nIndentationSize = 0;
	m_nNbMultipleSpaces = 0;
}

void CAppLogic::addIndentationSize(long n)
{
	m_nIndentationSize += n;
}

APICALL long CAppLogic::getIndentationSize()
{
	return m_nIndentationSize;
}


APICALL long CAppLogic::getNbMultipleSpaces()
{
	return m_nNbMultipleSpaces;
}





// content

void CAppLogic::initContentStats()
{
	m_nMinContentLength = 0x7FFFFFFF;
	m_nMaxContentLength = m_nTotalContentLength = m_nNbContentCells = 0;

	m_ndMinDuplicate = 1.0; // default 100% (to ensure that min(a,b) will work)
	m_ndMaxDuplicate = m_ndTotalDuplicate = m_ndTotalStdDevDuplicate = 0.0;

	m_nMultiplePartValues = 0;
}

APICALL long CAppLogic::getStructNbEmptyElementTags()
{
	XmlArrayNode elementCollection;

	getEmptyElementTags(elementCollection);

	return elementCollection.GetSize();
}

APICALL long CAppLogic::getNbEmptyElementTags() // in whole stream
{
	XmlArrayNode elementCollection;

	getEmptyElementTags(elementCollection);

	long nSize = elementCollection.GetSize();
	long nTotal = 0;
	for (long i=0;i<nSize;i++)
	{
		IXmlNode *p = elementCollection.GetAt(i);
		if (p)
			nTotal += p->getAmountOccurences(); // add occurences of this element in the whole stream
	}
	return nTotal;
}

void CAppLogic::getEmptyElementTags(/*out*/XmlArrayNode &elementCollection)
{
	getEmptyTags(m_pRoot, TRUE/*all elements*/, elementCollection);
}

APICALL long CAppLogic::getStructNbEmptyAttributeTags()
{
	XmlArrayNode attribCollection;

	getEmptyAttributeTags(attribCollection);

	return attribCollection.GetSize();
}

APICALL long CAppLogic::getNbEmptyAttributeTags() // in whole stream
{
	XmlArrayNode attribCollection;

	getEmptyAttributeTags(attribCollection);

	long nSize = attribCollection.GetSize();
	long nTotal = 0;
	for (long i=0;i<nSize;i++)
	{
		IXmlNode *p = attribCollection.GetAt(i);
		if (p)
			nTotal += p->getAmountOccurences(); // add occurences of this attribute in the whole stream
	}
	return nTotal;
}

APICALL void CAppLogic::getEmptyAttributeTags(/*out*/XmlArrayNode &attribCollection)
{
	getEmptyTags(m_pRoot, FALSE/*all attributes*/, attribCollection);
}

void CAppLogic::getEmptyTags(IXmlNode *p, BOOL bElementOnly, /*out*/XmlArrayNode &nodeCollection)
{

	if (!p)
		return;

	CElement *pElement = (CElement *)p;

	// add contribution of this element
	if (bElementOnly)
	{
		if ( pElement->getChildrenSize()==0 && pElement->getLOV().GetSize()==0 ) // no children, no content ==> that's an empty tag
			nodeCollection.Add( p );
	}
	else // bElementOnly = FALSE ==> process attributes instead
	{
		// process all attributes
		//
		long nAttributes = pElement->getAttributes().GetSize();
		for (long j=0; j<nAttributes; j++)
		{
			CAttribute *pAttribute = (CAttribute *) pElement->getAttributes().GetAt(j);
			if (pAttribute)
			{
				if ( pAttribute->getLOV().GetSize()==0 ) // no content ==> that's an empty tag
					nodeCollection.Add( pAttribute );
			}
		} // end process all attributes
	}

	// then process all direct element children
	//
	int nChildElementsSize = pElement->getChildrenSize();
	for (int k=0; k<nChildElementsSize; k++)
	{
		CElement *pChildElement = pElement->getChildrenElement(k);
		if (pChildElement)
			getEmptyTags(pChildElement,bElementOnly,nodeCollection);
	}

}


void CAppLogic::setMinContentLength(long nMin)
{
	m_nMinContentLength = nMin;
}

void CAppLogic::setMaxContentLength(long nMax)
{
	m_nMaxContentLength = nMax;
}

void CAppLogic::addContentLength(long nLength)
{
	m_nTotalContentLength += nLength;
	m_nNbContentCells++; // for avg computation
}

APICALL long CAppLogic::getMinContentLength()
{
	return m_nMinContentLength;
}

APICALL long CAppLogic::getMaxContentLength()
{
	return m_nMaxContentLength;
}

APICALL double CAppLogic::getMeanContentLength()
{
	return m_nNbContentCells ? double(m_nTotalContentLength)/m_nNbContentCells : 0.0;
}



APICALL double CAppLogic::getMinDuplicate() // %
{
	return m_ndMinDuplicate; // calculated by evalContentStatsMinMaxMeanDuplicates()
}
APICALL double CAppLogic::getMaxDuplicate() // %
{
	return m_ndMaxDuplicate; // calculated by evalContentStatsMinMaxMeanDuplicates()
}
APICALL double CAppLogic::getMeanDuplicate() // %
{
	if (!m_pRoot)
		return 0.0;

	long nTotalNodes = m_pRoot->getNbElementsAndAttributes() - 1; // -1 because __root__ should not be taken into account

	// m_nTotalDuplicate is calculated by evalContentStatsMinMaxMeanDuplicates()

	return nTotalNodes==0 ? 0.0 : (m_ndTotalDuplicate / nTotalNodes);
}
APICALL double CAppLogic::getStdDevDuplicate()
{
	if (!m_pRoot)
		return 0.0;

	evalContentStatsStdDevDuplicates( m_pRoot, getMeanDuplicate() );

	long nTotalNodes = m_pRoot->getNbElementsAndAttributes() - 1; // -1 because __root__ should not be taken into account

	// m_nTotalDuplicate is calculated by evalContentStatsMinMaxMeanDuplicates()

	return nTotalNodes==0 ? 0.0 : (m_ndTotalStdDevDuplicate / nTotalNodes);
}
void CAppLogic::addDuplicateValues(String *s)
{
	if (s==NULL)
		return;

	XMLNAMED *pBlock = XmlhashTableLookup(&m_hDuplicatedValues, s->GetData(), (void*)-1); // create XMLNAMED struct internally if new object
	if (!pBlock)
		pBlock = XmlhashTableLookup(&m_hDuplicatedValues, s->GetData(), (void*)-1); // again

	if (pBlock)
	{
		long nFrequency = long(pBlock->ptr)+1; // add an occurence to this value
		if (nFrequency==0) nFrequency++;
		pBlock->ptr = (void*)nFrequency;
	}
}
APICALL void CAppLogic::getTopDuplicateValues(long nNb, /*out*/XmlArrayString &stringCollection)
{
	XMLNAMED *pBlock = NULL;
	XMLNAMED *pBlock1 = NULL;
	XMLNAMED *pBlock2 = NULL;
	XMLNAMED *pBlock3 = NULL;

	long nRanking = -1;

	pBlock = XmlhashTableFirstItem(&m_hDuplicatedValuesIter,&m_hDuplicatedValues);
	if (pBlock==NULL) // empty hash table
		return;

	do
	{
		// pBlock is valid

		if ( long(pBlock->ptr) > nRanking )
		{
			nRanking = long(pBlock->ptr);
			pBlock1 = pBlock;
		}

	}
	while ( (pBlock = XmlhashTableNextItem(&m_hDuplicatedValuesIter,&m_hDuplicatedValues))!=NULL );

	if (pBlock1)
	{
		stringCollection.Add ( new String(pBlock1->name) );
	}
}

void CAppLogic::evalContentStatsMinMaxMeanDuplicates(IXmlNode *p)
{

	if (!p)
		return;

	CElement *pElement = (CElement *)p;

	// add contribution of this element
	double ndDuplicates = p->getDuplicatesInLOV();
	m_ndMinDuplicate = min(m_ndMinDuplicate, ndDuplicates);
	m_ndMaxDuplicate = max(m_ndMaxDuplicate, ndDuplicates);
	m_ndTotalDuplicate += ndDuplicates;

	// process all attributes
	//
	long nAttributes = pElement->getAttributes().GetSize();
	for (long j=0; j<nAttributes; j++)
	{
		CAttribute *pAttribute = (CAttribute *) pElement->getAttributes().GetAt(j);
		if (pAttribute)
		{
			double ndDuplicates = pAttribute->getDuplicatesInLOV();
			m_ndMinDuplicate = min(m_ndMinDuplicate, ndDuplicates);
			m_ndMaxDuplicate = max(m_ndMaxDuplicate, ndDuplicates);
			m_ndTotalDuplicate += ndDuplicates;
		}
	} // end process all attributes

	// then process all direct element children
	//
	int nChildElementsSize = pElement->getChildrenSize();
	for (int k=0; k<nChildElementsSize; k++)
	{
		CElement *pChildElement = pElement->getChildrenElement(k);
		if (pChildElement)
			evalContentStatsMinMaxMeanDuplicates(pChildElement);
	}

}


void CAppLogic::evalContentStatsStdDevDuplicates(IXmlNode *p, double ndMean)
{

	if (!p)
		return;

	CElement *pElement = (CElement *)p;

	// add contribution of this element
	m_ndTotalStdDevDuplicate += dabs(p->getDuplicatesInLOV(),ndMean);

	// process all attributes
	//
	long nAttributes = pElement->getAttributes().GetSize();
	for (long j=0; j<nAttributes; j++)
	{
		CAttribute *pAttribute = (CAttribute *) pElement->getAttributes().GetAt(j);
		if (pAttribute)
		{
			m_ndTotalStdDevDuplicate += dabs(pAttribute->getDuplicatesInLOV(),ndMean);
		}
	} // end process all attributes

	// then process all direct element children
	//
	int nChildElementsSize = pElement->getChildrenSize();
	for (int k=0; k<nChildElementsSize; k++)
	{
		CElement *pChildElement = pElement->getChildrenElement(k);
		if (pChildElement)
			evalContentStatsStdDevDuplicates(pChildElement,ndMean);
	}

}

void CAppLogic::addMultiplePartValues(long nIncr)
{
	m_nMultiplePartValues += nIncr;
}

APICALL double CAppLogic::getMultiplePartValuesRatio()
{
	long nbTotalElements = getStatsNbElements();

	return nbTotalElements==0 ? 0.0 : double(m_nMultiplePartValues)/nbTotalElements;
}


void CAppLogic::evalDatatypes(IXmlNode *p)
{

	if (!p)
		return;

	CElement *pElement = (CElement *)p;

	// add contribution of this element
	p->inferDatatype(	p->revealDatatype_firstvalues(), 
						p->revealDatatype_lastvalues(),
						p->revealDatatype_randomvalues() );

	// process all attributes
	//
	long nAttributes = pElement->getAttributes().GetSize();
	for (long j=0; j<nAttributes; j++)
	{
		CAttribute *pAttribute = (CAttribute *) pElement->getAttributes().GetAt(j);
		if (pAttribute)
		{
			pAttribute->inferDatatype(	pAttribute->revealDatatype_firstvalues(), 
										pAttribute->revealDatatype_lastvalues(),
										pAttribute->revealDatatype_randomvalues() );
		}
	} // end process all attributes

	// then process all direct element children
	//
	int nChildElementsSize = pElement->getChildrenSize();
	for (int k=0; k<nChildElementsSize; k++)
	{
		CElement *pChildElement = pElement->getChildrenElement(k);
		if (pChildElement)
			evalDatatypes(pChildElement);
	}

}















////////////////////////////////////////////////////////////////////////////////////////
//
//
//
// Callbacks and applogic
//
// constructs the hierarchy of elements and attributes
//
//


void flushContent(); // forward declaration


// callback handler called by the XML Parser anytime it reaches a new XML element
//
// userData : user data passed along the calls (see process())
// name : name of the current element
// atts[] : pairs of all attributes names/values
//
void cbk_startElement(void *userData, const char *name, const char **atts)
{
	flushContent();

	if (name!=NULL && strlen(name)!=0)
	{

		CElement *pParent = (CElement*)userData;
		
		if (pParent->getAppLogic())
		{
			pParent->getAppLogic()->addCurDepth(1);
			pParent->getAppLogic()->addStatsNbElements( 1 );
			if ( strchr(name,':')!=NULL )
				pParent->getAppLogic()->addStatsUseNamespaces( name );
		}


		if (pParent)
		{
			String szElementName = (char*) name;

			CElement *pKnownElement = pParent->getChildren(szElementName);	
			if (pKnownElement==NULL)
			{
				if (!(pKnownElement=pParent->addChildren(szElementName)))
				{
					pParent->getAppLogic()->setLastError( SAXML_ERROR_OUTOFMEMORY );
					trace2(TRC_ERROR,"CAppLogic::cbk_startElement",SAXML_ERRORSTRING_OUTOFMEMORY);
					return;
				}
			}

		

			// PREDICATE : pKnownElement not NULL

			// add attributes
			int i = 0;
			while (atts[i])
			{
				if ( !pKnownElement->checkAttributeOrder( (char*) atts[i], i>>1 ) )
					pParent->getAppLogic()->addStructStatsTotalAttribsDisorder(1);

				if ( strchr(atts[i],':')!=NULL )
					pParent->getAppLogic()->addStatsUseNamespaces( atts[i] );

				// catch attr name
				pKnownElement->addAttribute( (char*) atts[i], (char*) atts[i+1] );

				long len = strlen( atts[i+1] );
				long nMin = min( pParent->getAppLogic()->getMinContentLength(), len );
				pParent->getAppLogic()->setMinContentLength(nMin);
				long nMax = max( pParent->getAppLogic()->getMaxContentLength(), len );
				pParent->getAppLogic()->setMaxContentLength(nMax);
				pParent->getAppLogic()->addContentLength(len);

				i++;i++;

				pParent->getAppLogic()->addStatsNbAttributes( 1 );
			}

			pKnownElement->anotherOccurence();
			XML_SetUserData(pParent->getAppLogic()->getParser(), pKnownElement);

		}
		else
		{
			CAppLogic::setLastError( SAXML_ERROR_SAXCALLBACK );
			trace(TRC_ERROR,"Null CElement* passed in callback\n");
		}
	}

}


// callback handler called by the XML Parser anytime it reaches the end of an XML element
//
// userData : user data passed along the calls (see process())
// name : name of the current element
//
void cbk_endElement(void *userData, const char *name)
{
	flushContent();

	CElement *pCurrentElement = (CElement*)userData;
	if (pCurrentElement)
	{
		pCurrentElement->getAppLogic()->addCurDepth(-1);

		XML_SetUserData(pCurrentElement->getAppLogic()->getParser(), pCurrentElement->getParent());
	}
}



void cbk_contentElement(void *userData, const XML_Char *s, int len)
{
	// check out whether there is *real* content, other than 0D 0A and 20.
	if (s==NULL || len==0)
		return;

	CElement *pCurrentElement = (CElement*)userData;

	BOOL bContent=FALSE;

	int i;
	for (i=0; i<len && !bContent; i++)
	{
		bContent = (s[i]>0x20);
	}

	if (!bContent)
	{
		if (pCurrentElement)
			pCurrentElement->getAppLogic()->addIndentationSize(len);
		return;
	}

	if (pCurrentElement && pCurrentElement->getAppLogic()->isCDATASectionMode())
		pCurrentElement->getAppLogic()->addStatsCDATASections( 0, len );

	if (pCurrentElement)
		pCurrentElement->getAppLogic()->addIndentationSize(i);

	if (g_szContentLength==0) // simple malloc
	{
		long nRealLength = (len>16384) ? 16384 : len;
		strncpy(g_szContent,s,nRealLength);
		g_szContentLength = nRealLength;
	}
	else
	{
		if (g_szContentLength<16384)
		{
			long nRealLength = (g_szContentLength+len > 16384) ? 16384-g_szContentLength : len;
			
			strncpy(g_szContent+g_szContentLength,s,nRealLength);
			g_szContentLength += nRealLength;
		}
	}


	g_cpCurrentElement = (CElement*)userData;

}




void cbk_commenthandler(void *userData, const XML_Char *data)
{
	CElement *pParent = (CElement*)userData;
		
	if (pParent && pParent->getAppLogic())
		pParent->getAppLogic()->addStatsNbComments( 1, strlen(data) );
}

void cbk_cdataSection(void *userData)
{
	CElement *pParent = (CElement*)userData;
		
	if (pParent && pParent->getAppLogic())
	{
		pParent->getAppLogic()->addStatsCDATASections( 1, 0 /* size */ );
		pParent->getAppLogic()->setCDATASectionMode(TRUE);
	}
}

void cbk_endcdataSection(void *userData)
{
	CElement *pParent = (CElement*)userData;
		
	if (pParent && pParent->getAppLogic())
		pParent->getAppLogic()->setCDATASectionMode(FALSE);
}

void cbk_pi(void *userData, const XML_Char *target, const XML_Char *data)
{
	CElement *pParent = (CElement*)userData;
		
	if (pParent && pParent->getAppLogic())
		pParent->getAppLogic()->addStatsNbPI( 1 );
}




size_t ConvertUTF8ToISO88591(LPSTR inbuf, LPSTR outbuf); // forward declaration

void flushContent()
{
	if (g_szContentLength>0)
	{
		CElement *pCurrentElement = (CElement*) g_cpCurrentElement;
		if (pCurrentElement)
		{
			pCurrentElement->setHasContent( SAXML_CONTENT_YES );

			long len = g_szContentLength;
			char *s = g_szContent;

			long nMin = min( pCurrentElement->getAppLogic()->getMinContentLength(), len );
			pCurrentElement->getAppLogic()->setMinContentLength(nMin);
			long nMax = max( pCurrentElement->getAppLogic()->getMaxContentLength(), len );
			pCurrentElement->getAppLogic()->setMaxContentLength(nMax);
			pCurrentElement->getAppLogic()->addContentLength(len);

			if (len>0 && s!=NULL)
			{
				char *szTmp=new char[len+1];
				if (!szTmp)
				{
					pCurrentElement->getAppLogic()->setLastError( SAXML_ERROR_OUTOFMEMORY );
					trace2(TRC_ERROR,"CAppLogic::cbk_contentElement",SAXML_ERRORSTRING_OUTOFMEMORY);
					return;
				}
				strncpy(szTmp,s,len);
				szTmp[len]='\0';

				char *szTmpAscii=new char[len+1];
				szTmpAscii[0]='\0';

				ConvertUTF8ToISO88591(szTmp, szTmpAscii);

				delete [] szTmp; 

				String *st = new String(szTmpAscii);
				pCurrentElement->checkDuplicate( st );
				pCurrentElement->getLOV().Add( st );

				delete [] szTmpAscii;
			}

			// check out whether this content is in multiple parts
			if (pCurrentElement->getLOV().GetSize() != pCurrentElement->getAmountOccurences())
				pCurrentElement->getAppLogic()->addMultiplePartValues(1);

		}

		// make sure of no reentrance
		g_szContentLength = 0;
		g_szContent[0] = 0;
		g_cpCurrentElement = NULL;

	} // end 
}




long CAppLogic::countNbLines(char *buf, long nLength)
{
	long nblines = 0;

	if (buf)
	{
		long i=0;
		char c;
		while ( i<nLength && (c=buf[i++])!='\0' )
		{
			if (c==0x0A) // count all EOLs
				nblines++;
		}
	}

	return nblines;
}







void CAppLogic::dump()
{
	char szTemp[MAX_PATH], szT[MAX_PATH];
	long i,j;

	strcpy(szT,"(");
	strcat(szT,getStatsStreamIsFull()?"full":"portion only");
	strcat(szT,")");

	sprintf(szTemp,"\n\n\n");trace(TRC_INFO,szTemp);
	sprintf(szTemp,"-------- Statistical dump for (%s)--------\n\n",getStatsStreamFilename());trace(TRC_INFO,szTemp);
	long nStreamLength = getStatsStreamLength();
	sprintf(szTemp,"Stream size : %ld bytes %s\n", nStreamLength, szT);trace(TRC_INFO,szTemp);
	if (nStreamLength>0) {

	long nblines = getStatsNbLines() + ((nStreamLength==0)?0:1);
	long nbelements = getStatsNbElements();
	sprintf(szTemp,"Nb total lines : %ld\n", nblines);trace(TRC_INFO,szTemp); // +1 because there is always 1 line more than the amount of EOLs
	sprintf(szTemp,"Nb total elements : %ld\n", nbelements);trace(TRC_INFO,szTemp);
	sprintf(szTemp,"Nb total attributes : %ld\n", getStatsNbAttributes());trace(TRC_INFO,szTemp);
	sprintf(szTemp,"Nb total comments : %ld (%.2f%% of overall stream size)\n", 
					getStatsNbComments(),
					100*double(getStatsCommentsSize())/nStreamLength);trace(TRC_INFO,szTemp);
	sprintf(szTemp,"Nb total CDATA sections : %ld (%.2f%% of overall stream size)\n", 
					getStatsCDATASections(),
					100*double(getStatsCDATASectionsSize())/nStreamLength);trace(TRC_INFO,szTemp);
	sprintf(szTemp,"Nb total P. instructions : %ld\n", getStatsNbPI());trace(TRC_INFO,szTemp);
	sprintf(szTemp,"Nb total namespaces used : %ld\n", getStatsNbNamespaces());trace(TRC_INFO,szTemp);
	sprintf(szTemp,"\n");trace(TRC_INFO,szTemp);
	if (nbelements>0)
	{
		sprintf(szTemp,"-- Depth dump --\n\n");trace(TRC_INFO,szTemp);
		sprintf(szTemp,"Max : %ld\n", getDepthMax());trace(TRC_INFO,szTemp);
		sprintf(szTemp,"Mean : %.2f \n", getDepthMean()+1);trace(TRC_INFO,szTemp);
		sprintf(szTemp,"StdDev : %.2f\n", getDepthStdDev());trace(TRC_INFO,szTemp);
		for (i=1;i<5&&i<=getDepthMax();i++) // browse the 5 first depths
		{
			sprintf(szTemp,"Ratio of elements of depth %ld : %ld / %ld (%.2f%%)\n", 
						i, 
						getDepthOccurences(i),
						nbelements,
						100*double(getDepthOccurences(i))/nbelements);trace(TRC_INFO,szTemp);

			if ( getDepthOccurences(i) >0 ) // list all elements at this depth
			{
				strcpy(szT,"  ");
				XmlArrayNode elemcoll;
				getDepthElementCollection(i, elemcoll);
				long jn = elemcoll.GetSize();
				if (jn>0)
				{
					for (j=0;j<jn;j++)
					{
						strcat(szT,"<");
						IXmlNode *p = elemcoll.GetAt(j);
						if (p)
							strcat(szT,p->getName());
						strcat(szT,">");
					}
					strcat(szT,"\n");
					trace(TRC_INFO,szT);
				}
			}
		}
	}
	sprintf(szTemp,"\n");trace(TRC_INFO,szTemp);
	sprintf(szTemp,"-- Struct-only dump --\n\n");trace(TRC_INFO,szTemp);
	sprintf(szTemp,"Min nodename size \t: %ld bytes\n", getStructStatsMin());trace(TRC_INFO,szTemp);
	sprintf(szTemp,"Max nodename size \t: %ld bytes\n", getStructStatsMax());trace(TRC_INFO,szTemp);
	sprintf(szTemp,"Mean nodename size \t: %.2f bytes\n", getStructStatsMean());trace(TRC_INFO,szTemp);
	sprintf(szTemp,"StdDev nodename size \t: %.2f\n", getStructStatsStdDev());trace(TRC_INFO,szTemp);

	long nTotal = getStructStatsTotalOccurences();
	long nTotalSize = getStructStatsTotalSize(); if (nTotalSize==0) nTotalSize=1; // avoid GPF

	for (i=1;i<=getStructStatsMax();i++) // browse the 5 first depths
	{
		sprintf(szTemp,"Occurences of nodenames %ld-byte sized :\t %ld times \t(%ld bytes, %.2f%%)\n", 
			i, 
			getStructStatsNameOccurences(i),
			getStructStatsNameSizes(i),
			100*double(getStructStatsNameSizes(i))/nTotalSize);trace(TRC_INFO,szTemp);

		strcpy(szTemp,"   (");
		XmlArrayString namecoll;
		getStructStatsNamesCollection(i, namecoll);
		long jn = namecoll.GetSize();
		if (jn>0)
		{
			for (j=0;j<jn;j++)
			{
				if (j>0) strcat(szTemp,",");

				strcat(szTemp, namecoll.GetAt(j)->GetData() );
				
			}
			strcat(szTemp,")\n");
			trace(TRC_INFO,szTemp);

			namecoll.Destroy();
		}


	}
	
	
	sprintf(szTemp,"\nTotal size of structure : %ld bytes (%.2f%% of stream size)\n", 
			nTotalSize, 
			100*double(nTotalSize)/nStreamLength);trace(TRC_INFO,szTemp);
	

	// potential gain with reduced node sizes
	i = 0;
	nTotal = 0;
	evalStructStatsIfNodesSizesOptimized( getRootElement(), i, nTotal);

	if (nTotal>0)
	{
		sprintf(szTemp,"\n--> Potential gain on overall stream size : %.2f%%\n     obtained with struct size made of reduced nodename sizes : %ld bytes\n", 
			100*double(nStreamLength-(nStreamLength-nTotalSize+nTotal))/nStreamLength,
			nTotal);
		trace(TRC_INFO,szTemp);
	}

	// 
	sprintf(szTemp,"\n\n");trace(TRC_INFO,szTemp);
	sprintf(szTemp,"Min Nb Attributes per element : %ld\n", getStructStatsMinNbAttributsPerElement());trace(TRC_INFO,szTemp);
	sprintf(szTemp,"Max Nb Attributes per element : %ld\n", getStructStatsMaxNbAttributsPerElement());trace(TRC_INFO,szTemp);
	sprintf(szTemp,"Mean Nb Attributes per element : %.2f\n", getStructStatsMeanNbAttributsPerElement());trace(TRC_INFO,szTemp);
	sprintf(szTemp,"StdDev Nb Attributes per element : %.2f\n", getStructStatsStdDevNbAttributsPerElement());trace(TRC_INFO,szTemp);

	//
	sprintf(szTemp,"\n\n");trace(TRC_INFO,szTemp);
	sprintf(szTemp,"Attributes disorder ratio : %.2f%%\n", 100*getStructStatsAttribsDisorderRatio());trace(TRC_INFO,szTemp);
	

	// pattern stats
	sprintf(szTemp,"\n\n");trace(TRC_INFO,szTemp);
	sprintf(szTemp,"-- Pattern dump --\n\n");trace(TRC_INFO,szTemp);
	sprintf(szTemp,"Nb patterns : %ld\n", getNbPatterns());trace(TRC_INFO,szTemp);
	long nFlattenStreamLength = 0;
	if (getNbPatterns()==1) // only one dominant pattern in the stream has been detected
	{
		sprintf(szTemp,"Pattern : (%ld occurences)\n", getPatternOccurences());trace(TRC_INFO,szTemp);

		long nDefaultDepth= 0;
		showMostSeenPattern( getRootElement(), nDefaultDepth);

		sprintf(szTemp,"\nPattern height : %ld lines\n", getPatternHeight());trace(TRC_INFO,szTemp);
		long nPatternSize = getPatternSize();
		long nFlattenPatternSize = getPatternFlatSize();
		sprintf(szTemp,"Pattern size : %ld bytes\n", nPatternSize);
		trace(TRC_INFO,szTemp);
		sprintf(szTemp,"Flatten Pattern size : %ld bytes (%.2f%%)\n", 
				nFlattenPatternSize,
				100*double(nPatternSize-nFlattenPatternSize)/nPatternSize);
		trace(TRC_INFO,szTemp);

		nFlattenStreamLength = nStreamLength - ( nPatternSize-nFlattenPatternSize ) * getPatternOccurences();

		if (nFlattenStreamLength<nStreamLength)
		{
			sprintf(szTemp,"\n--> Potential gain on overall stream size : %.2f%% obtained with flatten patterns\n", 
				100*double(nStreamLength-nFlattenStreamLength)/nStreamLength);
			trace(TRC_INFO,szTemp);
		}
		else
		{
			sprintf(szTemp,"\n--> No potential gain by flattening patterns\n");
			trace(TRC_INFO,szTemp);
		}

	}

	// indentation portion
	sprintf(szTemp,"\n\n");trace(TRC_INFO,szTemp);
	sprintf(szTemp,"-- Indentation dump --\n\n");trace(TRC_INFO,szTemp);
	sprintf(szTemp,"Indentation size : %ld bytes (%.2f%% of stream size)\n", 
		getIndentationSize(),
		100*double(getIndentationSize())/nStreamLength);trace(TRC_INFO,szTemp);

	sprintf(szTemp,"\n--> Potential gain on overall stream size : %.2f%% obtained without indentation\n", 
		100*double(getIndentationSize())/nStreamLength);
	trace(TRC_INFO,szTemp);


	// white spaces portion
	sprintf(szTemp,"\n\n");trace(TRC_INFO,szTemp);
	sprintf(szTemp,"-- White spaces dump --\n\n");trace(TRC_INFO,szTemp);
	BOOL bSignificantWhiteSpaces = getNbMultipleSpaces() > ( nStreamLength*1/100 ); // exceeds 1% of total stream size
	if (bSignificantWhiteSpaces)
		sprintf(szTemp,"Nb multiples spaces between attributes : %ld bytes (%.2f%% of stream size)\n", 
		getNbMultipleSpaces(),
		100*double(getNbMultipleSpaces())/nStreamLength);
	else
		sprintf(szTemp,"Nb multiples spaces between attributes : not significant\n");
	trace(TRC_INFO,szTemp);



	// content portion
	sprintf(szTemp,"\n\n");trace(TRC_INFO,szTemp);
	sprintf(szTemp,"-- Content dump --\n\n");trace(TRC_INFO,szTemp);
	long nNbEmptyElementsInStruct = getStructNbEmptyElementTags();
	sprintf(szTemp,"Nb empty element tags in struct : %ld\n", nNbEmptyElementsInStruct);trace(TRC_INFO,szTemp);
	if ( nNbEmptyElementsInStruct>0 )
	{
		strcpy(szT,"  ");
		XmlArrayNode elemcoll;
		getEmptyElementTags(elemcoll);
		long jn = elemcoll.GetSize();
		if (jn>0)
		{
			for (j=0;j<jn;j++)
			{
				strcat(szT,"<");
				IXmlNode *p = elemcoll.GetAt(j);
				if (p)
					strcat(szT,p->getName());
				strcat(szT,">");
			}
			strcat(szT,"\n");
			trace(TRC_INFO,szT);
		}
	}
	long nNbEmptyElementsInStream = getNbEmptyElementTags();
	sprintf(szTemp,"Ratio of empty element tags in whole stream : %ld / %ld (%.2f%%)\n", 
						nNbEmptyElementsInStream,
						nbelements,
						100*double(nNbEmptyElementsInStream)/nbelements);trace(TRC_INFO,szTemp);
	sprintf(szTemp,"\n");trace(TRC_INFO,szTemp);

	long nNbEmptyAttribsInStruct = getStructNbEmptyAttributeTags();
	sprintf(szTemp,"Nb empty attribute tags in struct : %ld\n", nNbEmptyAttribsInStruct);trace(TRC_INFO,szTemp);
	if ( nNbEmptyAttribsInStruct>0 )
	{
		strcpy(szT,"   (");
		XmlArrayNode attribcoll;
		getEmptyAttributeTags(attribcoll);
		long jn = attribcoll.GetSize();
		if (jn>0)
		{
			for (j=0;j<jn;j++)
			{
				if (j>0)
					strcat(szT,",");

				IXmlNode *p = attribcoll.GetAt(j);
				if (p)
					strcat(szT,p->getName());
			}
			strcat(szT,")\n");
			trace(TRC_INFO,szT);
		}
	}
	if ( getStatsNbAttributes()>0 )
	{
		long nNbEmptyAttribsInStream = getNbEmptyAttributeTags();
		sprintf(szTemp,"Ratio of empty attribute tags in whole stream : %ld / %ld (%.2f%%)\n", 
						nNbEmptyAttribsInStream,
						getStatsNbAttributes(),
						100*double(nNbEmptyAttribsInStream)/getStatsNbAttributes());trace(TRC_INFO,szTemp);
	}
	sprintf(szTemp,"\n\n");trace(TRC_INFO,szTemp);

	sprintf(szTemp,"Min content size : %ld bytes\n", getMinContentLength());trace(TRC_INFO,szTemp);
	sprintf(szTemp,"Max content size : %ld bytes\n", getMaxContentLength());trace(TRC_INFO,szTemp);
	sprintf(szTemp,"Mean content size : %.2f bytes\n", getMeanContentLength());trace(TRC_INFO,szTemp);

	sprintf(szTemp,"\n\n");trace(TRC_INFO,szTemp);
	double nbMeanDuplicate = 100*getMeanDuplicate();
	szT[0]='\0';
	if (nbMeanDuplicate<10.0)
		strcpy(szT,"(content is highly uncorrelated)");
	else if (nbMeanDuplicate>=10.0 && nbMeanDuplicate<20.0)
		strcpy(szT,"(content is somewhat uncorrelated)");
	else if (nbMeanDuplicate>=20.0 && nbMeanDuplicate<35.0)
		strcpy(szT,"(content is somewhat correlated)");
	else if (nbMeanDuplicate>=35.0 && nbMeanDuplicate<50.0)
		strcpy(szT,"(content is correlated)");
	else if (nbMeanDuplicate>=50.0)
		strcpy(szT,"(content is highly correlated)");
	sprintf(szTemp,"Max content duplication ratio : %.2f%%\n", 100*getMaxDuplicate());trace(TRC_INFO,szTemp);
	sprintf(szTemp,"Mean content duplication ratio : %.2f%% %s\n", nbMeanDuplicate,szT);trace(TRC_INFO,szTemp);
	sprintf(szTemp,"StdDev content duplication : %.2f\n", 100*getStdDevDuplicate());trace(TRC_INFO,szTemp);

	if ( 100*getMaxDuplicate()>30.0) // at least 30% in one of the LOV is a threshold for us to show anything relevant
	{
		XmlArrayString stringcoll;
		getTopDuplicateValues(0,stringcoll);
		if (stringcoll.GetSize()>0)
		{
			String *s = stringcoll.GetAt(0);
			if (s)
				sprintf(szTemp,"Topmost duplicated content : \"%s\"\n", s->GetData());trace(TRC_INFO,szTemp);
		}
		stringcoll.Destroy();
	}

	sprintf(szTemp,"\n\n");trace(TRC_INFO,szTemp);
	double nbMultiplePartsValuesRatio = 100*getMultiplePartValuesRatio();
	szT[0]='\0';
	if (nbMultiplePartsValuesRatio<10.0)
		strcpy(szT,"(not significant)");
	else if (nbMultiplePartsValuesRatio>=10.0 && nbMultiplePartsValuesRatio<30.0)
		strcpy(szT,"(significant)");
	else if (nbMultiplePartsValuesRatio>=30.0)
		strcpy(szT,"(questionable)");
	sprintf(szTemp,"Multiple part values ratio : %.2f%% %s\n", nbMultiplePartsValuesRatio, szT);trace(TRC_INFO,szTemp);
	

	// namespaces
	sprintf(szTemp,"\n\n");trace(TRC_INFO,szTemp);
	XmlArrayString namecoll;
	getNamespaces(namecoll);
	if (namecoll.GetSize()>0)
	{
		sprintf(szTemp,"Use of namespaces ratio : %.2f%% (fraction of elements and attributes that use namespaces)\n", 100*getNamespacesUseRatio());trace(TRC_INFO,szTemp);
		sprintf(szTemp,"List of namespaces :\n");trace(TRC_INFO,szTemp);
		szT[0]='\0';
		long i;
		long nSize = namecoll.GetSize();
		for (i=0;i<nSize;i++)
		{
			if(szT[0]!='\0')
				strcat(szT,",");
	
			String *s = namecoll.GetAt(i);
			if (s)
			{
				strcat(szT, s->GetData() );
				strcat(szT,":"); // for decoration...
			}
		}
		strcpy(szTemp,"  ");
		strcat(szTemp,szT);
		trace(TRC_INFO,szTemp);
		namecoll.Destroy();
	}
	else
	{
		sprintf(szTemp,"Use of namespaces : none\n");trace(TRC_INFO,szTemp);
	}


	// cumulated gains
	//
	long nStructGain = nTotalSize+nTotal;
	long nPatternGain = nStreamLength - nFlattenStreamLength;
	long nIndentationGain = getIndentationSize();

	long nCumulatedGain = nStructGain + nPatternGain + nIndentationGain;

	trace(TRC_INFO,"\n\n");
	sprintf(szTemp,"\n--> CUMULATED GAIN ON OVERALL STREAM SIZE : %.2f%% (a factor of x %.1f, which is %ld bytes instead of %ld bytes)\n      obtained with reduced node sizes\n      and flatten patterns\n      and no indentation\n", 
		100*double(nCumulatedGain)/nStreamLength,
		1/(double(nStreamLength-nCumulatedGain)/nStreamLength),
		nStreamLength-nCumulatedGain,
		nStreamLength);
	trace(TRC_INFO,szTemp);

	} // end if (nStreamLength>0)




	trace(TRC_INFO,"\n\n");
}


void CAppLogic::clean()
{
	// destroy element hierarchy
	delete m_pRoot;
	m_pRoot = NULL;

	XmlhashTableDestroy(&m_hDuplicatedValues);

	m_arrNamespaces.Destroy();
}














//// -------------------------- UTF8 to Isolatin conversion --------------------------

/*
 * The UTF-FSS (aka UTF-2) encoding of UCS, as described in the
following
 * quote from Ken Thompson's utf-fss.c:
 *
 * Bits  Hex Min  Hex Max  Byte Sequence in Binary
 *   7  00000000 0000007f 0vvvvvvv
 *  11  00000080 000007FF 110vvvvv 10vvvvvv
 *  16  00000800 0000FFFF 1110vvvv 10vvvvvv 10vvvvvv
 *  21  00010000 001FFFFF 11110vvv 10vvvvvv 10vvvvvv 10vvvvvv
 *  26  00200000 03FFFFFF 111110vv 10vvvvvv 10vvvvvv 10vvvvvv 10vvvvvv
 *  31  04000000 7FFFFFFF 1111110v 10vvvvvv 10vvvvvv 10vvvvvv 10vvvvvv
10vvvvvv
 *
 * The UCS value is just the concatenation of the v bits in the
multibyte
 * encoding.  When there are multiple ways to encode a value, for
example
 * UCS 0, only the shortest encoding is legal.
 */

/* This implementation does not enforce the last restriction on input */

#define	CODE_ERROR	0x80
#ifndef REP_CHAR // replacement character
	#define REP_CHAR ' '
#endif

long utf8_getc(unsigned char ** pIn)
{
	int			c;
	long    	wc;
	int			extras;
	unsigned 	bit;

	if ((c = **pIn) == 0)
		return -1;
	(*pIn)++;
	if ((c & 0x80) == 0)		// ASCII character
		return c;
	if ((c & 0xc0) == 0x80)		// unexpected tail character 
		return CODE_ERROR;
	// how many extra bytes?
	extras = 1;
	for (bit = 0x20; (c & bit) != 0; bit >>= 1)
		extras++;
	if (extras > 5)
		return CODE_ERROR;
	// put all the bits together
	wc = c & (bit - 1);
	while (extras-- > 0) {
		if ((c = **pIn) == 0)
			return -1;
		(*pIn)++;
		if ((c & 0xc0) != 0x80) { // unexpected head character 
			(*pIn)--;
			return CODE_ERROR;
		} wc = (wc << 6) | (c & 0x3f);
	} return wc;
}

long utf8_putc(long wc, unsigned char ** pOut)
{
	long    tmp;
	int		extras;

	wc &= 0x7fffffffL;
	if ((wc & 0x7f) == wc) {  // ASCII character
		**pOut = (unsigned char)wc;
		(*pOut)++;
		return wc;
	}
	// how many extra bytes are required? 
	extras = 1;
	for (tmp = wc >> 11; tmp != 0; tmp >>= 5)
		extras++;
	// put header byte
	**pOut = ( 0xff & (0x1f80 >> extras) ) | (int)(wc >> (extras * 6));
	(*pOut)++;
	// put tail bytes
	while (extras-- != 0) {
		**pOut = 0x80 | ( 0x3f & (int)(wc >> (extras * 6)) );
		(*pOut)++;
	} return wc;
}





size_t ConvertUTF8ToISO88591(LPSTR inbuf, LPSTR outbuf)
{
	long	c;
	unsigned char * pIn;
	unsigned char * pOut;
	size_t unconverted = 0;
	pIn = (unsigned char *) inbuf;
	pOut = (unsigned char *) outbuf;
	while ((c = utf8_getc(&pIn)) != -1) {
		if (c < 0x100) {
           	*pOut++ = (unsigned char)c;
		}
		else {
			/* Character is not in destination character set, use 
			   replacement character */
           	*pOut++ = (unsigned char)REP_CHAR; 
			++unconverted;
		}
	}
	*pOut = 0;             /* Null terminate */
	return unconverted;
}


//// ---- end of 









///////////////////////////HASH CODE TABLE ////////////////////////////////////////////////


#define XMLINIT_SIZE 64

static int Xmlkeyeq(const char *s1, String &s2)
{
  return s2.Compare(s1);
}

static unsigned long Xmlhash(const char *s)
{
  unsigned long h = 0;
  while (*s)
    h = (h << 5) + h + (unsigned char)*s++;
  return h;
}

XMLNAMED *XmlhashTableLookup(XMLHASH_TABLE *table, const char *name, const void *ptr)
{
  size_t i;
  if (table->size == 0) // non initialized table
  {
    table->v = (XMLNAMED**) calloc(XMLINIT_SIZE, sizeof(XMLNAMED *));
    if (!table->v)
      return NULL;
    table->size = XMLINIT_SIZE;
    table->usedLim = XMLINIT_SIZE / 2;
    i = Xmlhash(name) & (table->size - 1);
  }
  else // find name in the table
  {
    unsigned long h = Xmlhash(name);
    for (i = h & (table->size - 1); table->v[i]; i == 0 ? i = table->size - 1 : --i) 
	{
      if (Xmlkeyeq(name, table->v[i]->name))
		return table->v[i]; // name found
    }

	// not found, prepare addition to the table
    if (table->used == table->usedLim) 
	{
      // check for overflow
      size_t newSize = table->size * 2;
      XMLNAMED **newV = (XMLNAMED**) calloc(newSize, sizeof(XMLNAMED *));
      if (!newV)
		return NULL;
      for (i = 0; i < table->size; i++)
		if (table->v[i]) 
		{
			size_t j;
			for (j = Xmlhash(table->v[i]->name) & (newSize - 1);
					newV[j];
					j == 0 ? j = newSize - 1 : --j)
				;

			newV[j] = table->v[i];
		}
      free(table->v);
      table->v = newV;
      table->size = newSize;
      table->usedLim = newSize/2;
      for (i = h & (table->size - 1);  table->v[i];  i == 0 ? i = table->size - 1 : --i)
			;
    }
  }

  if (ptr) // new item added to the table
  {
	table->v[i] = new XMLNAMED(); // was (in original code) table->v[i] = (XMLNAMED*) calloc(1, sizeof(XMLNAMED) );
    if (!table->v[i])
      return 0;

	table->v[i]->name = (char*) name; // name is const char*

    table->v[i]->ptr = (void*)ptr;
    (table->used)++;
  }
  return NULL; //was return table->v[i];
}

void XmlhashTableDestroy(XMLHASH_TABLE *table)
{
  for (size_t i = 0; i < table->size; i++) 
  {
    delete table->v[i];
  }

  if (table->v)
	free(table->v);

  // init again
  XmlhashTableInit(table);

}

void XmlhashTableInit(XMLHASH_TABLE *p)
{
  p->size = 0;
  p->usedLim = 0;
  p->used = 0;
  p->v = NULL;
}

long XmlhashTableSize(XMLHASH_TABLE *p)
{
	return p? p->used : 0;
}


XMLNAMED *XmlhashTableFirstItem(XMLHASH_TABLE_ITER *iter, XMLHASH_TABLE *p)
{
	if (!iter)
		return NULL;

	iter->p = p->v;
	iter->end = iter->p + p->size;

	return XmlhashTableNextItem(iter, p);
}

XMLNAMED *XmlhashTableNextItem(XMLHASH_TABLE_ITER *iter, XMLHASH_TABLE *p)
{
	while (iter->p != iter->end) 
	{
		XMLNAMED *tem = *(iter->p)++;
		if (tem)
			return tem;
	}
	return NULL;
}



