

#ifndef _APPLOGIC
#define _APPLOGIC

#include "stdtrace.h"
#include "XmlIdDef.h"
#include "Templates.h"
#include "..\xmlparser\xmlparse.h"
#include "Element.h"


#include "IXmlStats.h"




// callback functions used by SAX.
void cbk_startElement(void *userData, const XML_Char *name, const XML_Char **atts);
void cbk_endElement(void *userData, const XML_Char *name);
void cbk_contentElement(void *userData, const XML_Char *s, int len);
void cbk_commenthandler(void *userData, const XML_Char *data);
void cbk_cdataSection(void *userData);
void cbk_endcdataSection(void *userData);
void cbk_pi(void *userData, const XML_Char *target, const XML_Char *data);

class CElement; // forward declaration
class CXmlFileManager; // forward declaration

////////////////////////////////////////////////////////////////////////////////////////
//
//
// CAppLogic class
//
//
/////////////////////////////////////////////////////////////////////////////////////////
// Stephane Rodriguez, 1st july 2002
//
//
class CAppLogic : public IXmlStats
{
	// Members
	//
protected:
	CElement			*m_pRoot;

	XML_Parser			m_parser;
	BOOL				m_bExternalForcedStop;
	String				m_szFilename;

	void*				m_cpParent; // used to send message (SetParamInt, ...)

	BOOL				m_bProcessFinished;
	CXmlFileManager*	m_pResourceManager;

	static long			m_nLastError;


	// general stats
	long				m_nStreamLength;
	BOOL				m_bStreamFull;
	long				m_nStreamNbLines;
	long				m_nNbElements, m_nNbAttributes, m_nNbComments, m_nNbCDATASections, m_nNbPI;
	long				m_nCommentSize, m_nCDATASectionsSize;
	BOOL				m_bCDATASectionMode; // used in character cbk to know if we are parsing CDATA section content
	long				m_nUseNamespaces;
	XmlArrayString		m_arrNamespaces;

	// depth stats
	long				m_nCurDepth;
	XmlArrayInt			m_arrDepthOccurences;

	// struct stats
	XmlArrayInt			m_arrStructNameOccurences;
	XmlArrayInt			m_arrStructNameSizes;
	long				m_nMinNbAttribsPerElement, m_nMaxNbAttribsPerElement, m_nTotalNbAttribsPerElement;
	double				m_ndStdDevNbAttribsPerElement;
	long				m_nTotalAttribsDisorder;

	// pattern stats
	long				m_nPatternElementMaxOccurences, m_nbPatterns;

	// indentation stats
	long				m_nIndentationSize, m_nNbMultipleSpaces;

	// content stats
	long				m_nMinContentLength, m_nMaxContentLength, m_nTotalContentLength, m_nNbContentCells;
	double				m_ndMinDuplicate, m_ndMaxDuplicate, m_ndTotalDuplicate, m_ndTotalStdDevDuplicate;
	XMLHASH_TABLE		m_hDuplicatedValues;
	XMLHASH_TABLE_ITER	m_hDuplicatedValuesIter;
	long				m_nMultiplePartValues; // used to reflect if values in element are in one piece or several pieces

	// Constructor
	//
public:
	CAppLogic();
	CAppLogic(const CAppLogic &obj);
	virtual ~CAppLogic();

	virtual void init();

public:
	// IUnknown implementation
	virtual HRESULT __stdcall	QueryInterface(const IID& iid, void** ppv) ;			
	virtual ULONG __stdcall		AddRef() ;
	virtual ULONG __stdcall		Release() ;

	// Accessors
	//
	virtual CElement*			__stdcall getRootElement() { return m_pRoot; }
	virtual XML_Parser&			__stdcall getParser() { return m_parser; }

	// Methods
	//


	// general stats
	void initGeneralStats();
	void setStatsStreamIsFull(BOOL bFull);
	void addStatsStreamBufferSize(long nBufferSize);
	void addStatsStreamNbLines(long nIncr);
	void addStatsNbElements(long nIncr);
	void addStatsNbAttributes(long nIncr);
	void addStatsNbComments(long nIncr, long nSize); // one or more comment, size of comment in bytes
	void addStatsCDATASections(long nIncr, long nSize);
	void setCDATASectionMode(BOOL bActivate=TRUE);
	BOOL isCDATASectionMode();
	void addStatsNbPI(long nIncr);
	APICALL char *getStatsStreamFilename();
	APICALL long getStatsStreamLength();
	APICALL BOOL getStatsStreamIsFull();
	APICALL long getStatsNbLines();
	APICALL long getStatsNbElements();
	APICALL long getStatsNbAttributes();
	APICALL long getStatsNbComments();
	APICALL long getStatsCommentsSize();
	APICALL long getStatsCDATASections();
	APICALL long getStatsCDATASectionsSize();
	APICALL long getStatsNbPI();

	// depth
	void initDepthStats();
	void setCurDepth(long n);
	void addCurDepth(long nIncr);
	APICALL long getDepthMax(); // 1-based
	APICALL double getDepthMean(); // 1-based
	APICALL double getDepthStdDev();
	APICALL long getDepthOccurences(long nDepth); // depth is 1-based, depth should be given in the interval [1,..., getDepthMax()]
	APICALL void getDepthElementCollection(long nDepth, /*out*/XmlArrayNode &elementCollection);

	// struct-only
	void initStructStats();
	void setStructNameOccurences(long n, long nOccurences);
	void setStructNameSizes(long n, long nSize); // total size in bytes within struct of "element with size n"
	APICALL long getStructStatsTotalOccurences();
	APICALL long getStructStatsTotalSize();
	APICALL long getStructStatsMin();
	APICALL long getStructStatsMax();
	APICALL double getStructStatsMean();
	APICALL double getStructStatsStdDev();
	APICALL long getStructStatsNameOccurences(long i); // i is the length in bytes of the requested elem+attrib info, i is in the interval [1,2,...,getStructStatsMax()]
	APICALL long getStructStatsNameSizes(long i); // i is the length in bytes of the requested elem+attrib info, i is in the interval [1,2,...,getStructStatsMax()]
	APICALL void getStructStatsNamesCollection(long i, /*out*/XmlArrayString &nameCollection); // all elements+attribs which name is of length i bytes
	APICALL long getStructStatsMinNbAttributsPerElement();
	APICALL long getStructStatsMaxNbAttributsPerElement();
	APICALL double getStructStatsMeanNbAttributsPerElement();
	APICALL double getStructStatsStdDevNbAttributsPerElement();
	void evalStructStatsIfNodesSizesOptimized(IXmlNode *p, /*out*/long &i, /*out*/long &nTotal);
	void addStructStatsTotalAttribsDisorder(long nIncr);
	APICALL double getStructStatsAttribsDisorderRatio();
	void addStatsUseNamespaces(const char *szName);
	APICALL double getNamespacesUseRatio(); // %
	APICALL long getStatsNbNamespaces();
	APICALL void getNamespaces(XmlArrayString &namecoll); // clone known namespaces collection

	// pattern
	void initPatternStats();
	APICALL long getNbPatterns();
	APICALL long getPatternOccurences(); // how many times the pattern appears in the stream
	APICALL long getPatternHeight(); // height in nb of lines of the pattern
	APICALL long getPatternSize(); // size in bytes of the struct pattern (data not taken into account)
	APICALL long getPatternFlatSize(); // size in bytes of the flattened struct pattern (data not taken into account)

	// indentation
	void initIndentationStats();
	void addIndentationSize(long n);
	APICALL long getIndentationSize();

	// misc stats
	APICALL long getNbMultipleSpaces(); // in whole stream

	// content
	void initContentStats();
	APICALL long getStructNbEmptyElementTags();
	APICALL long getNbEmptyElementTags(); // in whole stream
	APICALL void getEmptyElementTags(/*out*/XmlArrayNode &elementCollection);
	APICALL long getStructNbEmptyAttributeTags();
	APICALL long getNbEmptyAttributeTags(); // in whole stream
	APICALL void getEmptyAttributeTags(/*out*/XmlArrayNode &attribCollection);
	void setMinContentLength(long nMin);
	void setMaxContentLength(long nMax);
	void addContentLength(long nLength);
	APICALL long getMinContentLength();
	APICALL long getMaxContentLength();
	APICALL double getMeanContentLength();

	APICALL double getMinDuplicate(); // %, max over all LOVs
	APICALL double getMaxDuplicate(); // %, max over all LOVs
	APICALL double getMeanDuplicate(); // %, max over all LOVs
	APICALL double getStdDevDuplicate();
	void addDuplicateValues(String *s);
	APICALL void getTopDuplicateValues(long nNb, /*out*/XmlArrayString &stringCollection);

	void addMultiplePartValues(long nIncr);
	APICALL double getMultiplePartValuesRatio();

	// parse the XML stream (file, http://, ...)
	virtual BOOL __stdcall process(char *szFilename);
	virtual void __stdcall forceStop(); // command to stop processing inside ::process

	// generate a debug log of the internal structure
	virtual void __stdcall dump();

	// drop the internal structure
	virtual void __stdcall clean();


	virtual void  setParent(void *pParent);
	virtual void* getParent();

	static void setLastError(long nError) { m_nLastError = nError; }
	static long getLastError() { return m_nLastError; }


	// Internal
	//
protected:
	long countNbLines(char *buf, long nLength);
	void evalStructStatsMinMaxMean(IXmlNode *p);
	void evalStructStatsMinMaxMeanAttribsPerElement(IXmlNode *p);
	void evalStructStatsStdDevAttribsPerElement(IXmlNode *p, double ndMean);
	void evalMostSeenPattern(IXmlNode *p, /*in*/long &nThreshold);
	void findMostSeenPattern(IXmlNode *p, long &nNbPatterns);
	void showMostSeenPattern(IXmlNode *p, long nCurDepth);
	void getRawPatternSize(IXmlNode *p, /*in*/BOOL bFlatten, /*out*/long &nSize, /*out*/long &nHeight);
	void getDepthElementCollection(IXmlNode *p, long nTargetDepth, long nCurDepth, /*out*/XmlArrayNode &elementCollection);
	void getStructStatsNamesCollection(IXmlNode *p, /*size of names in bytes*/long i, /*out*/XmlArrayString &nameCollection); // all elements+attribs which name is of length i bytes
	void getEmptyTags(IXmlNode *p, BOOL bElementOnly, /*out*/XmlArrayNode &nodeCollection);
	void evalContentStatsMinMaxMeanDuplicates(IXmlNode *p);
	void evalContentStatsStdDevDuplicates(IXmlNode *p, double ndMean);
	void evalDatatypes(IXmlNode *p);

	// Prohibited use
	//
private:
	CAppLogic& operator=(const CAppLogic &obj);

private:
	long m_cRef; // object life cycle

};










#endif