
#ifndef _PERSONALSTRING_
#define _PERSONALSTRING_

#include <stdio.h>
#include <string.h>


// --------------------------------------------------------------------------------
//
//
//	String class : exposes the equivalent of the MFC API, without the MFC.
//
//
//
class String
{
	// Members
protected:
	char *m_str;

	// Constructor/Destructor
public:
	String() 
	{ 
		m_str = NULL;
	}
	String(char *szInputSource) 
	{ 
		m_str = NULL;
		if (szInputSource)
			Copy(szInputSource);
	}
	String(const String &szInputSource) 
	{ 
		m_str = NULL;
		Copy(szInputSource.GetData());
	}
	~String()
	{
		Empty();
	}

// Methods
public:
	void Empty()
	{
		if (m_str)
		{
			free( m_str ); 
			m_str = NULL;
		}
	}

	long GetLength() const
	{ 
		return m_str ? strlen(m_str) : 0; 
	}

	bool IsEmpty() const
	{
		return GetLength()==0;
	}

	char *GetData() const
	{
		return (char*) m_str;
	}

	operator char*()
	{
		return GetData();
	}

	char GetAt(long nIndex)
	{
		return nIndex<GetLength() ? m_str[nIndex] : 0;
	}

	char RawGetAt(long nIndex) // no sanity check, use only in performance issues
	{
		return m_str[nIndex];
	}

	char operator[](long nIndex)
	{
		return GetAt(nIndex);
	}

	void SetAt(long nIndex, char ch)
	{
		if (nIndex<GetLength())
			m_str[nIndex] = ch;
	}

	void RawSetAt(long nIndex, char ch) // no sanity check, use only in performance issues
	{
		m_str[nIndex] = ch;
	}

	void Copy(char *szInputSource)
	{
		Empty();

		long len = strlen(szInputSource);
		if (szInputSource && len>0)
		{
			m_str = (char*) malloc(len+1);
			if (m_str) strcpy(m_str, szInputSource);
		}
	}

	void Copy(const String &szInputSource)
	{
		Copy( szInputSource.GetData() );
	}

	void operator =(char *szInputSource)
	{
		Copy( szInputSource );
	}

	void operator =(const String &szInputSource)
	{
		Copy( szInputSource );
	}

	void ToLower()
	{
		if (GetLength()>0)
			strlwr(m_str);
	}

	void ToUpper()
	{
		if (GetLength()>0)
			strupr(m_str);
	}

	bool Compare(const char *szCompareString)
	{
		return strcmp(GetData(),szCompareString)==0;
	}

	bool Compare(String &szCompareString)
	{
		return strcmp(GetData(),szCompareString.GetData())==0;
	}

	bool CompareNoCase(String &szCompareString)
	{
		return stricmp(GetData(),szCompareString.GetData())==0;
	}

	bool operator ==(String &szCompareString)
	{
		return Compare(szCompareString);
	}

	long Find(long nStartIndex, char *szStringToFind) // find string
	{
		if (szStringToFind && nStartIndex<GetLength())
		{
			char *szResult = strstr(m_str+nStartIndex, szStringToFind);
			return szResult ? (long)(szResult - m_str) : -1;
		}
		else
			return -1;
	}

	long Find(long nStartIndex, char sCharToFind) // find char
	{
		char szTemp[2];
		szTemp[0] = sCharToFind;
		szTemp[1] = '\0';

		return Find(nStartIndex, szTemp);
	}

	long FindNoCase(long nStartIndex, char *szStringToFind) // find string (without case)
	{
		if (!szStringToFind)
			return -1;

		String szTempLower = GetData();
		szTempLower.ToLower();
		String szStringToFindLower = szStringToFind;
		szStringToFindLower.ToLower();

		return szTempLower.Find(nStartIndex, szStringToFindLower.GetData());
	}

	long FindNoCase(long nStartIndex, char szCharToFind) // find char (without case)
	{
		char szTemp[2];
		szTemp[0] = szCharToFind;
		szTemp[1] = '\0';

		return FindNoCase(nStartIndex, szTemp);
	}

	long ReverseFind(long nStartIndex, char ch)
	{
		if (nStartIndex<GetLength())
		{
			char *szResult = strrchr(m_str+nStartIndex, ch);
			return szResult ? (long)(szResult - m_str) : -1;
		}
		else
			return -1;
	}

	String Left(long nCount)
	{
		String temp;
		char *szTemp=(char*) malloc(nCount+1);
		szTemp[0]='\0';
		if (nCount>0 && GetLength()>0)
		{
			strncat(szTemp, m_str, nCount>GetLength() ? GetLength() : nCount);
			strcat(szTemp, "\0");
		}
		temp = szTemp;
		free(szTemp);
		return temp;
	}

	String ExcludeLeft(long nCount) // remove first nCount chars
	{
		return Right(GetLength()-nCount);
	}

	String Mid(long nIndex, long nCount)
	{
		String temp;
		char *szTemp=(char*) malloc(nCount+1);
		szTemp[0]='\0';
		if (nIndex>=0 && nCount>0 && GetLength()>0)
		{
			strncat(szTemp, m_str+nIndex, (nCount+nIndex)>GetLength() ? GetLength()-nIndex : nCount);
			strcat(szTemp, "\0");
		}
		temp = szTemp;
		free(szTemp);
		return temp;
	}

	String Right(long nCount)
	{
		return Mid(GetLength()-nCount, nCount);
	}

	void InternalLeft(long nCount) //  applies to *this* object
	{
		String t;
		t = Left(nCount);
		*this = t;
	}

	void InternalExcludeLeft(long nCount) // remove first nCount chars,  applies to *this* object
	{
		InternalRight(GetLength()-nCount);
	}

	void InternalMid(long nIndex, long nCount) //  applies to *this* object
	{
		String t;
		t = Mid(nIndex,nCount);
		*this = t;
	}

	void InternalRight(long nCount) //  applies to *this* object
	{
		InternalMid(GetLength()-nCount, nCount);
	}




	void Append(char *szAdditionalString)
	{
		if (!m_str)
			Copy(szAdditionalString);
		else
		{
			long n = strlen(szAdditionalString);
			if (n>0)
			{
				m_str = (char*) realloc (m_str, GetLength()+n );
				strcat(GetData(),szAdditionalString);
			}
		}
	}

	void Append(String &szAdditionalString)
	{
		if (!szAdditionalString.IsEmpty())
			Append( szAdditionalString.GetData() );
	}

	void Append(char sAdditionalChar)
	{
		char szAdditionalString[2];
		szAdditionalString[0] = sAdditionalChar;
		szAdditionalString[1] = '\0';
		Append(szAdditionalString);
	}

	void operator +=(String &szAdditionalString)
	{
		Append(szAdditionalString);
	}

	void operator +=(char *szAdditionalString)
	{
		Append(szAdditionalString);
	}

	void operator +=(char sAdditionalChar)
	{
		Append(sAdditionalChar);
	}

	String operator +(String &szAdditionalString)
	{
		String szTemp = *this;
		szTemp.Append(szAdditionalString);
		return szTemp;
	}

	String operator +(char *szAdditionalString)
	{
		String szTemp = *this;
		szTemp.Append(szAdditionalString);
		return szTemp;
	}

	String operator +(char sAdditionalChar)
	{
		String szTemp = *this;
		szTemp.Append(sAdditionalChar);
		return szTemp;
	}

	void TrimLeft(char ch=' ')
	{
		if (IsEmpty())
			return;
		
		long nSize = GetLength();
		long nAmount = 0;
		bool bContinue = true;
		for (long i=0; i<nSize && bContinue; i++)
		{
			if (RawGetAt(i)==ch)
				nAmount++;
			else
				bContinue = false;
		}

		if (!bContinue || nAmount>0)
			InternalExcludeLeft( nAmount );
	}

	void TrimRight(char ch=' ')
	{
		if (IsEmpty())
			return;
		
		long nSize = GetLength();
		long nAmount = 0;
		bool bContinue = true;
		for (long i=0; i<nSize && bContinue; i++)
		{
			if (RawGetAt(nSize-1-i)==ch)
				nAmount++;
			else
				bContinue = false;
		}

		if (!bContinue || nAmount>0)
			InternalLeft( GetLength()-nAmount );
	}

	void Trim(char ch=' ')
	{
		TrimLeft(ch);
		TrimRight(ch);
	}

	long GetNumber()
	{
		if (IsEmpty())
			return 0L;

		long longValue = 0;
		char *c = strpbrk(GetData(), "0123456789.,-+");
		if (c) sscanf(c, "%ld", &longValue); 
		return longValue;
	}

	float GetFloat()
	{
		if (IsEmpty())
			return 0.0f;

		float floatValue = 0;
		char *c = strpbrk(GetData(), "0123456789.,-+");
		if (c) sscanf(c, "%f", &floatValue); 
		return floatValue;
	}


	void ReplaceChar(char ch, char sReplacementChar)
	{
		if (IsEmpty())
			return;

		long nSize = GetLength();
		long nIndex=0;
		long nIndex2;
		while ( nIndex<nSize && (nIndex2=Find(nIndex,ch))>-1 )
		{
			RawSetAt( nIndex2, sReplacementChar);

			nIndex = ++nIndex2; // ensure no infinite loop trap
			
			// loop
		}
	}

	void ReplaceCharWithString(char ch, char *szReplacementString) // szReplacementString==NULL allowed
	{
		if (IsEmpty())
			return;

		String szTemp;
		long nSize = GetLength();
		long nIndex=0;
		long nIndex2;
		while ( nIndex<nSize && (nIndex2=Find(nIndex,ch))>-1 )
		{
			// copy first portion of string
			if (nIndex2>nIndex)
			{
				szTemp += Mid(nIndex, nIndex2-nIndex);
			}

			// then copy replacement string
			if (szReplacementString)
			{
				szTemp += szReplacementString;
			}

			nIndex = ++nIndex2; // ensure no infinite loop trap
			
			// loop
		}

		// don't forget to copy the remainder of the input string, if any
		if (nIndex<nSize)
			szTemp += ExcludeLeft(nIndex);

		*this = szTemp;
	}

	void ReplaceCharWithString(char ch, String &szReplacementString)
	{
		ReplaceCharWithString(ch, szReplacementString.GetData() );
	}


}; 

typedef String &RString;

#endif