
#include "stdafx.h"
#include "XmlIdDef.h"
#include "Element.h"
#include "AppLogic.h"
#include "IXmlStats.h"


////////////////////////////////////////////////////////////////////////////////////////
//
//
// IXmlNode class
//
//
/////////////////////////////////////////////////////////////////////////////////////////
// Stephane Rodriguez, 1st july 2002
//
//


IXmlNode::IXmlNode()
{
	m_nHasContent=SAXML_CONTENT_NO;
	setSort();
	setDatatype();
	setOccurences(0);
	setMostSeen(FALSE);
	m_nDuplicatesInLOV = 0;
	setAppLogic(NULL);
}

IXmlNode::~IXmlNode() 
{
	cleanLOVValues();
}


inline long IXmlNode::getXmlElementType() 
{ 
	return m_nXmlElementType; 
}

inline void IXmlNode::setName(char *szName) 
{ 
	if (!szName)
		return;

	m_szName = (strlen(szName)>0)? szName : "__root__"; 
}

inline char *IXmlNode::getName() 
{ 
	return m_szName.GetData(); 
}


inline void IXmlNode::setRenameTo(char *szName) 
{ 
	if (!szName)
		return;

	m_szRenameTo = (strlen(szName)>0)? szName : "__root__"; 
}

inline char *IXmlNode::getRenameTo() 
{ 
	return m_szRenameTo.GetData(); 
}


inline void IXmlNode::setSort(long nSort)
{
	m_nSort = nSort;
}

inline long IXmlNode::getSort()
{
	return m_nSort;
}


inline void IXmlNode::setDatatype(long nDatatype) // for example, SAXML_DATATYPE_STRING
{
	m_nDatatype = nDatatype;
}

inline long IXmlNode::getDatatype()
{
	return m_nDatatype;
}

inline char *IXmlNode::getDatatypeAsString()
{
	switch ( getDatatype() )
	{
		case SAXML_DATATYPE_STRING : return ""; break;
		case SAXML_DATATYPE_NUMBER : return "number"; break;
		case SAXML_DATATYPE_FLOAT : return "float"; break;
		case SAXML_DATATYPE_DATE : return "date"; break;
		case SAXML_DATATYPE_TIME : return "time"; break;
		case SAXML_DATATYPE_CURRENCY : return "currency"; break;
		case SAXML_DATATYPE_URL : return "url"; break;
		case SAXML_DATATYPE_EMAIL : return "email"; break;
	}
	return NULL;
}

inline void IXmlNode::anotherOccurence() 
{ 
	m_nOccurences++; 
}

inline void IXmlNode::setOccurences(long nOccurences)
{
	m_nOccurences = nOccurences;
}

inline long IXmlNode::getAmountOccurences() 
{ 
	return m_nOccurences; 
}

// managing PCDATA section
inline void IXmlNode::setHasContent(long nHasContent) 
{ 
	// assign new value only if we haven't assigned SAXML_CONTENT_YES before
	if (m_nHasContent==SAXML_CONTENT_NO || m_nHasContent==SAXML_CONTENT_NOTAVAILABLE)
		m_nHasContent = nHasContent; 
}

inline long IXmlNode::getHasContent() 
{ 
	return m_nHasContent; 
}




// managing the fact that this element belongs to the most seen pattern in the stream
inline void IXmlNode::setMostSeen(BOOL bMostSeen)
{
	m_bMostSeenPattern = bMostSeen;
}

inline BOOL IXmlNode::isMostSeen()
{
	return m_bMostSeenPattern;
}



// managing hierarchy
inline BOOL IXmlNode::isParent(IXmlNode *pNode)
{
	return FALSE;
}



// managing LOV
XmlArrayString &IXmlNode::getLOV()
{
	return m_arrLOV;
}

void IXmlNode::cleanLOVValues()
{
	long i;

	int size = m_arrLOV.GetSize();
	for (i=0; i<size; i++)
	{
		delete m_arrLOV.GetAt(i); // String *
	}
	m_arrLOV.RemoveAll();

	int sizeD = m_DuplicatesValues.GetSize();
	for (i=0; i<sizeD; i++)
	{
		delete m_DuplicatesValues.GetAt(i); // String *
	}
	m_DuplicatesValues.RemoveAll();

}

BOOL IXmlNode::checkDuplicate(String *s) // check if string s is already known in the LOV values
{
	int size = m_arrLOV.GetSize();
	if (size==0 || s==NULL)
		return FALSE;

	// list from bottom to top ; limit listing to a window only of values for performance reasons
	BOOL bDuplicate = FALSE;
	for (long i=size-1 ; i>-1 && i>size-1-SAXML_WINDOWSAMPLEVALUES && !bDuplicate; i--)
	{
		String *st=m_arrLOV.GetAt(i);
		if (st)
			bDuplicate = strcmp( st->GetData(), s->GetData())==0;
	}

	if (bDuplicate)
	{
		m_nDuplicatesInLOV++;

		m_DuplicatesValues.Add( new String(s->GetData())  );

		if (m_cpManager)
			m_cpManager->addDuplicateValues( s );
	}

	return bDuplicate;
}


double IXmlNode::getDuplicatesInLOV() // in % (between 0 and 1)
{
	return m_arrLOV.GetSize()==0 ? 0.0 : (double(m_nDuplicatesInLOV) / m_arrLOV.GetSize()) ;
}


String *IXmlNode::getMostDuplicateInLOV()
{
	// rank the values by counting how many times they appear in the whole LOV
	long nSize = m_DuplicatesValues.GetSize();
	if (nSize==0)
		return NULL; // nothing to do here

	long nRanking = -1;
	String *s=NULL;

	for (long i=0; i<nSize; i++)
	{
		long nCnt = 0;

		String *scandidate = m_DuplicatesValues.GetAt(i);
		if (scandidate)
		{
			long nSizeLOV = m_arrLOV.GetSize();
			for (long j=0;j<nSizeLOV;j++)
			{
				String *st=m_arrLOV.GetAt(j);
				if (st && strcmp(scandidate->GetData(), st->GetData())==0)
					nCnt++;
			}

		} // end if (scandidate)

		if (nCnt>nRanking) // at this moment, the candidate string is the most seen of all duplicate strings
		{
			s = scandidate;
			nRanking = nCnt;
		}
	} // end for i

	return s;
}

void IXmlNode::cleanOccurences()
{
	setOccurences(0);
}


void IXmlNode::setAppLogic(CAppLogic *p)
{
	m_cpManager = p;
}

CAppLogic *IXmlNode::getAppLogic()
{
	return m_cpManager;
}






char *_mystrspnp( const char * _cpc1, const char * _cpc2) { return (*(_cpc1 += strspn(_cpc1,_cpc2))!='\0') ? (char*)_cpc1 : NULL; }

BOOL hasDigits(char *szValue, char excludedchar)
{
	if (*szValue=='\0')
		return FALSE;

	register char c;
	char *s=szValue;
	while (*s!='\0')
	{
		c = *s;
		s++;
		if ((c<'0' || c>'9') && !(c==excludedchar))
			return FALSE;
	}

	return TRUE;
}

BOOL checkDatatypeFloat(char *szValue)
{
	if (*szValue=='\0')
		return FALSE;

	register char c;
	char *s=szValue;
	while (*s!='\0')
	{
		c = *s;
		s++;
		if ((c<'0' || c>'9') && !((c=='.') || (c=='e') || (c=='-') || (c=='+')))
			return FALSE;
	}

// This fucking code does not do what it should because of a fucking strspnp C-runtime function
//	if ( (strpbrk(szValue,"0123456789")!=NULL)  && (_mystrspnp(szValue,"0123456789.e-+")!=szValue))
//		return TRUE;

	return TRUE;
}

BOOL checkDatatypeLong(char *szValue)
{
	if (*szValue=='\0')
		return FALSE;

	register char c;
	char *s=szValue;
	while (*s!='\0')
	{
		c = *s;
		s++;
		if ((c<'0' || c>'9') && !((c=='-') || (c=='+')))
			return FALSE;
	}

	return TRUE;

// This fucking code does not do what it should because of a fucking strspnp C-runtime function
	// note : strspnp returns the pointer to the fist character in szValue that doesn't belong to the list "0123456789-+"
//	return _mystrspnp(szValue,"0123456789-+") ? FALSE : TRUE ;
}

BOOL checkDatatypeDate(char *szValue)
{
	int nbSlash = 0;
	char *s = szValue;
	while ( (s=strchr(s,'/')) ) // find amount of / chars
	{
		s++; // next char please
		nbSlash++;
	}
	if (nbSlash>1 && hasDigits(szValue,'/')) // looks like a fucking date
		return TRUE;

	int nbDoubleDot = 0;
	s = szValue;
	while ( (s=strchr(s,':')) ) // find amount of : chars
	{
		s++; // next char please
		nbDoubleDot++;
	}
	if (nbDoubleDot>1 && hasDigits(szValue,':')) // looks like a fucking date
		return TRUE;

	return FALSE;
}

BOOL checkDatatypeTime(char *szValue)
{
	return FALSE;
}

BOOL checkDatatypeDateTime(char *szValue)
{
	return FALSE;
}

BOOL checkDatatypeCurrency(char *szValue)
{
	if (strchr(szValue,'$') || strchr(szValue,'') || strchr(szValue,''))
		return TRUE;

	return FALSE;
}

BOOL checkDatatypeUrl(char *szValue)
{
	if (strnicmp(szValue,"http://",7)==0)
		return TRUE;

	return FALSE;
}

BOOL checkDatatypeEmail(char *szValue)
{
	if (strchr(szValue,'@'))
		return TRUE;

	return FALSE;
}


// managing internal data type
// algorithm1 : check first values in the LOV
int IXmlNode::revealDatatype_firstvalues()
{
	long nSize = m_arrLOV.GetSize();
	if (nSize==0)
		return SAXML_DATATYPE_STRING; // default datatype

	long nbTypes = 8; 

	XmlArrayInt typecoll;
	typecoll.SetSize(nbTypes);

	BOOL bRes;

	for (long i=0;i<nSize && i<SAXML_WINDOWSAMPLEVALUES/2; i++)
	{
		String *s= m_arrLOV.GetAt(i);
		if (!s) continue;

		char *szValue = s->GetData();

		bRes = 	checkDatatypeFloat(szValue);
		if (bRes)
		{
			bRes = checkDatatypeLong(szValue);
			if (bRes)
				typecoll.SetAt(1, typecoll.GetAt(1)+1 ); // [1] for long
			else
				typecoll.SetAt(0, typecoll.GetAt(0)+1 ); // [0] for float
		}
		else // not a float, check out whether it is a date
		{
			bRes = checkDatatypeDate(szValue);
			if (bRes)
			{
				typecoll.SetAt(2, typecoll.GetAt(2)+1 ); // [2] for general date

				// TODO : check out TIME and DATE+TIME
			}
			else // not a date/time, check out whether it is a currency
			{
				bRes = checkDatatypeCurrency(szValue);
				if (bRes)
					typecoll.SetAt(5, typecoll.GetAt(5)+1 ); // [5] for currency
				else
				{
					bRes = checkDatatypeUrl(szValue);
					if (bRes)
						typecoll.SetAt(6, typecoll.GetAt(6)+1 ); // [6] for general HTTP URL
					else
					{
						bRes = checkDatatypeEmail(szValue);
						if (bRes)
							typecoll.SetAt(7, typecoll.GetAt(7)+1 ); // [7] for email address
					}
				}
			}
		}
	}

	// sort typecoll and extract the most elevant datatype from it
	long nMax = 0;
	long iCandidate = -1;

	for (i=0;i<nbTypes;i++)
	{
		if (typecoll.GetAt(i) > nMax)
		{
			nMax = typecoll.GetAt(i);
			iCandidate = i;
		}
	}

	return (iCandidate==0) ? SAXML_DATATYPE_FLOAT :
			(iCandidate==1) ? SAXML_DATATYPE_NUMBER :
			 (iCandidate==2) ? SAXML_DATATYPE_DATE :
			  (iCandidate==3) ? SAXML_DATATYPE_TIME :
			   (iCandidate==4) ? SAXML_DATATYPE_DATE :
				(iCandidate==5) ? SAXML_DATATYPE_CURRENCY : 
				 (iCandidate==6) ? SAXML_DATATYPE_URL :
				  (iCandidate==7) ? SAXML_DATATYPE_EMAIL :SAXML_DATATYPE_STRING;
}
// algorithm2 : check last values in the LOV
int IXmlNode::revealDatatype_lastvalues()
{
	long nSize = m_arrLOV.GetSize();
	if (nSize==0)
		return SAXML_DATATYPE_STRING; // default datatype

	long nbTypes = 8; 

	XmlArrayInt typecoll;
	typecoll.SetSize(nbTypes);

	BOOL bRes;

	for (long i=nSize-1;i>=0 && i>nSize-(SAXML_WINDOWSAMPLEVALUES/2); i--)
	{
		String *s= m_arrLOV.GetAt(i);
		if (!s) continue;

		char *szValue = s->GetData();

		bRes = 	checkDatatypeFloat(szValue);
		if (bRes)
		{
			bRes = checkDatatypeLong(szValue);
			if (bRes)
				typecoll.SetAt(1, typecoll.GetAt(1)+1 ); // [1] for long
			else
				typecoll.SetAt(0, typecoll.GetAt(0)+1 ); // [0] for float
		}
		else // not a float, check out whether it is a date
		{
			bRes = checkDatatypeDate(szValue);
			if (bRes)
			{
				typecoll.SetAt(2, typecoll.GetAt(2)+1 ); // [2] for general date

				// TODO : check out TIME and DATE+TIME
			}
			else // not a date/time, check out whether it is a currency
			{
				bRes = checkDatatypeCurrency(szValue);
				if (bRes)
					typecoll.SetAt(5, typecoll.GetAt(5)+1 ); // [5] for currency
				else
				{
					bRes = checkDatatypeUrl(szValue);
					if (bRes)
						typecoll.SetAt(6, typecoll.GetAt(6)+1 ); // [6] for general HTTP URL
					else
					{
						bRes = checkDatatypeEmail(szValue);
						if (bRes)
							typecoll.SetAt(7, typecoll.GetAt(7)+1 ); // [7] for email address
					}
				}
			}
		}
	}

	// sort typecoll and extract the most elevant datatype from it
	long nMax = 0;
	long iCandidate = -1;

	for (i=0;i<nbTypes;i++)
	{
		if (typecoll.GetAt(i) > nMax)
		{
			nMax = typecoll.GetAt(i);
			iCandidate = i;
		}
	}

	return (iCandidate==0) ? SAXML_DATATYPE_FLOAT :
			(iCandidate==1) ? SAXML_DATATYPE_NUMBER :
			 (iCandidate==2) ? SAXML_DATATYPE_DATE :
			  (iCandidate==3) ? SAXML_DATATYPE_TIME :
			   (iCandidate==4) ? SAXML_DATATYPE_DATE :
				(iCandidate==5) ? SAXML_DATATYPE_CURRENCY : 
				 (iCandidate==6) ? SAXML_DATATYPE_URL :
				  (iCandidate==7) ? SAXML_DATATYPE_EMAIL :SAXML_DATATYPE_STRING;

}
// algorithm3 : check several random values in the LOV
int IXmlNode::revealDatatype_randomvalues()
{
	long nSize = m_arrLOV.GetSize();
	if (nSize==0)
		return SAXML_DATATYPE_STRING; // default datatype

	long nbTypes = 8; 

	XmlArrayInt typecoll;
	typecoll.SetSize(nbTypes);

	BOOL bRes;

	long i;
	for (long j=0;j<SAXML_WINDOWSAMPLEVALUES/2; j++)
	{
		i = long( double(rand())*nSize / RAND_MAX ); // get a random index from [0, ..., nSize-1]
		if (i==nSize) i=0; // avoid stupid GPF

		String *s= m_arrLOV.GetAt(i);
		if (!s) continue;

		char *szValue = s->GetData();

		bRes = 	checkDatatypeFloat(szValue);
		if (bRes)
		{
			bRes = checkDatatypeLong(szValue);
			if (bRes)
				typecoll.SetAt(1, typecoll.GetAt(1)+1 ); // [1] for long
			else
				typecoll.SetAt(0, typecoll.GetAt(0)+1 ); // [0] for float
		}
		else // not a float, check out whether it is a date
		{
			bRes = checkDatatypeDate(szValue);
			if (bRes)
			{
				typecoll.SetAt(2, typecoll.GetAt(2)+1 ); // [2] for general date

				// TODO : check out TIME and DATE+TIME
			}
			else // not a date/time, check out whether it is a currency
			{
				bRes = checkDatatypeCurrency(szValue);
				if (bRes)
					typecoll.SetAt(5, typecoll.GetAt(5)+1 ); // [5] for currency
				else
				{
					bRes = checkDatatypeUrl(szValue);
					if (bRes)
						typecoll.SetAt(6, typecoll.GetAt(6)+1 ); // [6] for general HTTP URL
					else
					{
						bRes = checkDatatypeEmail(szValue);
						if (bRes)
							typecoll.SetAt(7, typecoll.GetAt(7)+1 ); // [7] for email address
					}
				}
			}
		}
	}

	// sort typecoll and extract the most elevant datatype from it
	long nMax = 0;
	long iCandidate = -1;

	for (i=0;i<nbTypes;i++)
	{
		if (typecoll.GetAt(i) > nMax)
		{
			nMax = typecoll.GetAt(i);
			iCandidate = i;
		}
	}

	return (iCandidate==0) ? SAXML_DATATYPE_FLOAT :
			(iCandidate==1) ? SAXML_DATATYPE_NUMBER :
			 (iCandidate==2) ? SAXML_DATATYPE_DATE :
			  (iCandidate==3) ? SAXML_DATATYPE_TIME :
			   (iCandidate==4) ? SAXML_DATATYPE_DATE :
				(iCandidate==5) ? SAXML_DATATYPE_CURRENCY : 
				 (iCandidate==6) ? SAXML_DATATYPE_URL :
				  (iCandidate==7) ? SAXML_DATATYPE_EMAIL :SAXML_DATATYPE_STRING;

}
// recoup everything here
void IXmlNode::inferDatatype(int nFromFirstValues, int nFromLastValues, int nFromRandomValues)
{
	setDatatype(SAXML_DATATYPE_STRING); // default.

	if ((nFromFirstValues==nFromLastValues) || (nFromLastValues==nFromRandomValues) || (nFromFirstValues==nFromRandomValues))
	{
		if (nFromFirstValues==nFromLastValues)
			setDatatype(nFromFirstValues);
		else if (nFromFirstValues==nFromRandomValues)
			setDatatype(nFromFirstValues);
		else if (nFromLastValues==nFromRandomValues)
			setDatatype(nFromLastValues);
	}
}








////////////////////////////////////////////////////////////////////////////////////////
//
//
// CAttribute class : represents an XML attribute 
//
//
/////////////////////////////////////////////////////////////////////////////////////////
// Stephane Rodriguez, 1st july 2002
//
//
CAttribute::CAttribute() 
{
	m_nXmlElementType = SAXML_ATTRIBUTELEVEL;

	setAssociatedElement(NULL);
}

CAttribute::~CAttribute() 
{
}


void CAttribute::setAssociatedElement(IXmlNode *pNode)
{
	m_cpAssociatedNode = pNode;
}
IXmlNode *CAttribute::getAssociatedElement()
{
	return m_cpAssociatedNode;
}








////////////////////////////////////////////////////////////////////////////////////////
//
//
// CElement class : represents an XML element 
//
//
/////////////////////////////////////////////////////////////////////////////////////////
// Stephane Rodriguez, 1st july 2002
//
//

CElement::CElement( CAppLogic *ptr, CElement *pParent, char *szName ) 
{ 
	m_nXmlElementType = SAXML_ELEMENTLEVEL;
	setAppLogic(ptr); 
	m_pParent = pParent; 
	setName(szName); 
	setRenameTo(szName);

	XmlhashTableInit(&m_Children);
}

CElement::~CElement() 
{
	long size = m_arrAttributes.GetSize(),i;
	for (i=0; i<size; i++)
		delete (CAttribute*) m_arrAttributes.GetAt(i);

	size = getChildrenSize();
	for (i=0; i<size; i++)
		delete getChildrenElement(i);

	XmlhashTableDestroy(&m_Children);
}



CElement& CElement::operator=(const CElement &obj)
{
	// prohibited use
	return *this;
}



// Methods
//

// managing type for all attributes
BOOL CElement::addAttribute(char *szAttributeName, char *szAttributeValue, 
							char *szRenameTo, 
							int nAttributeDatatype,
							int nOccurences,
							int nSort,
							int nHasContent)
{
	BOOL bResult = FALSE;

	BOOL bFound = FALSE;
	int i = 0;
	CAttribute *pAttrib = NULL;
	while (!bFound && i<m_arrAttributes.GetSize())
	{
		pAttrib = (CAttribute *) m_arrAttributes.GetAt(i);
		if (pAttrib && pAttrib->getName())
			bFound = !strcmp(pAttrib->getName(), szAttributeName);
		i++;
	}

	if (bFound)
	{
		// just add the attr value
		if (szAttributeValue && strlen(szAttributeValue)>0)
		{
			String *s = new String( szAttributeValue );

			pAttrib->checkDuplicate( s );
			pAttrib->getLOV().Add( s );

			pAttrib->setHasContent( SAXML_CONTENT_YES );
		}
		else
			pAttrib->setHasContent( nHasContent );

		pAttrib->anotherOccurence();

		return TRUE;
	}

	CAttribute *pNewAttrib = new CAttribute();
	if (pNewAttrib)
	{
		pNewAttrib->setName(szAttributeName);

		if (szRenameTo==NULL || strlen(szRenameTo)==0)
			pNewAttrib->setRenameTo(szAttributeName);
		else
			pNewAttrib->setRenameTo(szRenameTo);

		pNewAttrib->setAppLogic( getAppLogic() );

		pNewAttrib->setSort(nSort);

		pNewAttrib->setDatatype(nAttributeDatatype);

		pNewAttrib->setOccurences(nOccurences);

		m_arrAttributes.Add( pNewAttrib );

		// just add the attr value
		if (szAttributeValue && strlen(szAttributeValue)>0)
		{
			String *s = new String( szAttributeValue );

			pNewAttrib->checkDuplicate( s );
			pNewAttrib->getLOV().Add( s );

			pNewAttrib->setHasContent( SAXML_CONTENT_YES );
		}
		else
			pNewAttrib->setHasContent( nHasContent );


		pNewAttrib->setAssociatedElement(this);

		bResult = TRUE;
	}
	else
	{
		trace2(TRC_ERROR,"CElement::addAttribute",SAXML_ERRORSTRING_OUTOFMEMORY);
		bResult = FALSE;
	}

	return bResult;
}



// managing element hierarchy
void CElement::setParent(CElement *pParent)
{
	m_pParent = pParent;
}

CElement *CElement::getParent() 
{ 
	return m_pParent; 
}

long CElement::getChildrenSize()
{
	return XmlhashTableSize(&m_Children);
}

CElement *CElement::addChildren(String &szElementName)
{
	CElement *pNewElement = new CElement( getAppLogic(), this, (char*)szElementName );
	if (pNewElement) 
	{
		XMLNAMED *pBlock = XmlhashTableLookup(&m_Children, szElementName, pNewElement);

		// also add the element to an ordered array
		m_arrOrderedElements.Add( pNewElement );
	}

	return pNewElement;
}

void CElement::addChildren(CElement *pChildren)
{
	if (!pChildren)
		return;

	XMLNAMED *pBlock = XmlhashTableLookup(&m_Children, pChildren->getName(), pChildren);
}


CElement *CElement::getChildren(String &szName)
{
	if (szName.IsEmpty())
		return NULL;

	CElement *pElement=NULL;

	// Find an element in the table using hash keys : direct access almost granted
	XMLNAMED *pBlock = (XMLNAMED*) XmlhashTableLookup(&m_Children, szName.GetData(), NULL);
	if (pBlock)
		pElement = (CElement*) pBlock->ptr;

	return pElement;
}

CElement *CElement::getChildrenElement(long i)
{
	XMLNAMED *pBlock = NULL;

	pBlock = (i==0) ?	XmlhashTableFirstItem(&m_ChildrenIter,&m_Children) :
						XmlhashTableNextItem(&m_ChildrenIter,&m_Children);

	return pBlock? (CElement*) pBlock->ptr : NULL;
}

char *CElement::getChildrenName(long i)
{
	XMLNAMED *pBlock = NULL;

	pBlock = (i==0) ?	XmlhashTableFirstItem(&m_ChildrenIter,&m_Children) :
						XmlhashTableNextItem(&m_ChildrenIter,&m_Children);

	return pBlock? pBlock->name : NULL;
}

BOOL CElement::checkElementOrder(CElement *pChildren, long nOrder) // returns TRUE if element does not exist or if it is in the right order ; FALSE otherwise
{
	if (!pChildren)
		return TRUE;

	BOOL bFound = FALSE;
	long i = 0;
	CElement *p = NULL;
	while (!bFound && i<m_arrOrderedElements.GetSize())
	{
		p = (CElement *) m_arrOrderedElements.GetAt(i);
		bFound = p && (p==pChildren);
		i++;
	}

	if (!bFound)
		return TRUE; // does not exist yet

	i--;

	return (nOrder==i);

}



// managing attributes
XmlArrayLPVoid &CElement::getAttributes() 
{ 
	return m_arrAttributes; 
}

CAttribute *CElement::getAttribute(char *szAttrName) // 
{
	if (!szAttrName || strlen(szAttrName)==0)
		return NULL;

	BOOL bFound = FALSE;
	int i = 0;
	CAttribute *pAttrib = NULL;
	while (!bFound && i<m_arrAttributes.GetSize())
	{
		pAttrib = (CAttribute *) m_arrAttributes.GetAt(i);
		if (pAttrib && pAttrib->getName())
			bFound = !strcmp(pAttrib->getName(), szAttrName);
		i++;
	}
	return bFound ? pAttrib : NULL;
}


long CElement::getAmountAttributes() 
{ 
	return m_arrAttributes.GetSize(); 
}

BOOL CElement::checkAttributeOrder(char *szAttrName, long nOrder) // returns TRUE if attrib does not exist or if it is in the right order ; FALSE otherwise
{
	if (!szAttrName || strlen(szAttrName)==0)
		return TRUE;

	BOOL bFound = FALSE;
	long i = 0;
	CAttribute *pAttrib = NULL;
	while (!bFound && i<m_arrAttributes.GetSize())
	{
		pAttrib = (CAttribute *) m_arrAttributes.GetAt(i);
		if (pAttrib && pAttrib->getName())
			bFound = !strcmp(pAttrib->getName(), szAttrName);
		i++;
	}

	if (!bFound)
		return TRUE; // does not exist yet

	i--;

	return (nOrder==i);
}



// managing hierarchy
BOOL CElement::isParent(IXmlNode *pNode)
{
	BOOL bResult = FALSE;

	// check whether element pNode is a children of this element.
	// note : children is not necessarily seen as direct children
	// note : pNode is either an attribute or an element

	if (!pNode)
		return FALSE;

	char szNodeName[SAXML_URLMAXLENGTH];
	strcpy(szNodeName, pNode->getName());

	// pNode can be one of our attributes
	if (getAttribute(szNodeName))
		return TRUE;

	// pNode can be one of our direct children
	long i = 0;
	CElement *pElem = NULL;
	while (!bResult && i<getChildrenSize())
	{
		pElem = getChildrenElement(i);
		if (pElem)
		{
			bResult = !strcmp(pElem->getName(), szNodeName);

			if (!bResult)
				bResult = pElem->isParent(pNode);
		}

		i++;
	}

	return bResult;
}


long CElement::getNbElements() // whole hierarchy
{
	long nbElementsChildren = 0;

	int nChildElementsSize = getChildrenSize();
	for (int k=0; k<nChildElementsSize; k++)
	{
		CElement *pChildElement = getChildrenElement(k);
		if (pChildElement)
			nbElementsChildren += pChildElement->getNbElements();
	}

	return 1+nbElementsChildren;
}

long CElement::getNbElementsAndAttributes() // whole hierarchy
{
	long nbElementsChildren = 0;

	int nChildElementsSize = getChildrenSize();
	for (int k=0; k<nChildElementsSize; k++)
	{
		CElement *pChildElement = getChildrenElement(k);
		if (pChildElement)
			nbElementsChildren += pChildElement->getNbElementsAndAttributes();
	}

	return getAmountAttributes()+1+nbElementsChildren;
}



void CElement::cleanLOVValues()
{
	IXmlNode::cleanLOVValues();

	// process all attributes
	//
	long nAttributes = getAttributes().GetSize();
	for (long j=0; j<nAttributes; j++)
	{
		CAttribute *pAttribute = (CAttribute *) getAttributes().GetAt(j);
		if (pAttribute)
			pAttribute->cleanLOVValues();
	} // end process all attributes

	// then process all direct element children
	//
	int nChildElementsSize = getChildrenSize();
	for (int k=0; k<nChildElementsSize; k++)
	{
		CElement *pChildElement = getChildrenElement(k);
		if (pChildElement)
			pChildElement->cleanLOVValues();
	}
}


void CElement::cleanOccurences()
{
	IXmlNode::cleanOccurences();

	// process all attributes
	//
	long nAttributes = getAttributes().GetSize();
	for (long j=0; j<nAttributes; j++)
	{
		CAttribute *pAttribute = (CAttribute *) getAttributes().GetAt(j);
		if (pAttribute)
			pAttribute->cleanOccurences();
	} // end process all attributes

	// then process all direct element children
	//
	int nChildElementsSize = getChildrenSize();
	for (int k=0; k<nChildElementsSize; k++)
	{
		CElement *pChildElement = getChildrenElement(k);
		if (pChildElement)
			pChildElement->cleanOccurences();
	}
}

