
#ifndef _DPXMLELEMENT_
#define _DPXMLELEMENT_

#include "Templates.h"


class CAppLogic; // forward declaration



////////////////////////////////////////////////////////////////////////////////////////
//
//
// IXmlNode class
//
//
/////////////////////////////////////////////////////////////////////////////////////////
// Stephane Rodriguez, 1st july 2002
//
//

class IXmlNode // interface to all XML elements, including ELEMENTS and ATTRIBUTES
{

	// Members
	//
protected:
	long				m_nXmlElementType;

	String				m_szName;
	String				m_szRenameTo;
	long				m_nSort;
	long				m_nDatatype;
	long				m_nOccurences;
	long				m_nHasContent;
	BOOL				m_bSeenOnce;
	BOOL				m_bMostSeenPattern; // TRUE is this node belongs to the most seen pattern
	long				m_nDuplicatesInLOV; // ratio of same values in LOV

	XmlArrayString		m_arrLOV;
	XmlArrayString		m_DuplicatesValues;

	CAppLogic			*m_cpManager; // copy ptr

	// constructor/destructor
	//
public:
	IXmlNode();
	~IXmlNode();


	// Methods
	//
public:
	virtual long getXmlElementType(); // to know whether this node is an ELEMENT or an ATTRIBUTE

	virtual void setName(char *szName);
	virtual char *getName();

	virtual void setRenameTo(char *szName);
	virtual char *getRenameTo();

	virtual void setSort(long nSort=SAXML_NOSORT);
	virtual long getSort();

	virtual void setDatatype(long nDatatype=SAXML_DATATYPE_STRING);
	virtual long getDatatype();
	virtual char *getDatatypeAsString();

	// managing occurences
	virtual void anotherOccurence();
	virtual void setOccurences(long nOccurences);
	virtual long getAmountOccurences();

	// managing PCDATA section
	virtual void setHasContent(long nHasContent=SAXML_CONTENT_YES);
	virtual long getHasContent();

	// managing the fact that this element belongs to the most seen pattern in the stream
	virtual void setMostSeen(BOOL bMostSeen=TRUE);
	virtual BOOL isMostSeen();

	// managing hierarchy
	virtual BOOL isParent(IXmlNode *pNode); // override method if necessary

	// managing LOV
	virtual XmlArrayString &getLOV();
	virtual void cleanLOVValues();
	// small API for duplicates in LOV
	virtual BOOL checkDuplicate(String *s); // check if string s is already known in the LOV
	virtual double getDuplicatesInLOV(); // in % (between 0 and 1)
	virtual String *getMostDuplicateInLOV();

	virtual void cleanOccurences();
	

	// managing internal data type
	// algorithm1 : check first values in the LOV
	virtual int revealDatatype_firstvalues();
	// algorithm2 : check last values in the LOV
	virtual int revealDatatype_lastvalues();
	// algorithm3 : check several random values in the LOV
	virtual int revealDatatype_randomvalues();
	// recoup everything here
	virtual void inferDatatype(int nFromFirstValues, int nFromLastValues, int nFromRandomValues);


	void setAppLogic(CAppLogic *p);
	CAppLogic *getAppLogic();

};




////////////////////////////////////////////////////////////////////////////////////////
//
//
// CAttribute class : represents an XML attribute 
//
//
/////////////////////////////////////////////////////////////////////////////////////////
// Stephane Rodriguez, 1st july 2002
//
//
class CAttribute : public IXmlNode
{

	// Members
	//
protected:
	IXmlNode *m_cpAssociatedNode; // copy ptr

	// constructor/destructor
	//
public:
	CAttribute();
	virtual ~CAttribute();

	// Methods
	//
	void setAssociatedElement(IXmlNode *pNode);
	IXmlNode *getAssociatedElement();

	// Prohibited use
	//
private:
	CAttribute& operator=(const CAttribute &obj) { return *this; }


};





////////////////////////////////////////////////////////////////////////////////////////
//
//
// CElement class : represents an XML element 
//
//
/////////////////////////////////////////////////////////////////////////////////////////
// Stephane Rodriguez, 1st july 2002
//
//
class CElement : public IXmlNode
{
	// Members
	//
protected:
	XMLHASH_TABLE		m_Children; // direct children
	XMLHASH_TABLE_ITER	m_ChildrenIter;

	XmlArrayLPVoid		m_arrAttributes;
	XmlArrayLPVoid		m_arrOrderedElements;

	CElement			*m_pParent;


public:
	// constructor 
	//
	CElement( CAppLogic *ptr, CElement *pParent, char *szName ); 
	virtual ~CElement();

	// Methods
	//



	// managing type for all attributes
	BOOL addAttribute(char *szAttributeName, char *szAttributeValue, 
						char *szRenameTo=NULL, 
						int nAttributeDatatype=SAXML_DATATYPE_STRING, 
						int nOccurences=1,
						int nSort=SAXML_NOSORT,
						int nHasContent=SAXML_CONTENT_NO);


	// managing element hierarchy
	void				setParent(CElement *pParent);
	CElement			*getParent();
 
	CElement			*addChildren(String &szElementName);
	void				addChildren(CElement *pChildren);

	long				getChildrenSize();
	CElement			*getChildren(String &szName);
	CElement			*getChildrenElement(long i);
	char				*getChildrenName(long i);
	BOOL				checkElementOrder(CElement *pChildren, long nOrder); // returns TRUE if element does not exist or if it is in the right order ; FALSE otherwise

	// managing attributes
	XmlArrayLPVoid		&getAttributes(); // set of attributes
	CAttribute			*getAttribute(char *szAttrName); // 
	long				getAmountAttributes();
	BOOL				checkAttributeOrder(char *szAttrName, long nOrder); // returns TRUE if attrib does not exist or if it is in the right order ; FALSE otherwise

	// managing hierarchy
	BOOL				isParent(IXmlNode *pNode);

	long				getNbElements(); // whole hierarchy
	long				getNbElementsAndAttributes(); // whole hierarchy


	void				cleanLOVValues();
	void				cleanOccurences();

	// Prohibited use
	//
private:
	CElement& operator=(const CElement &obj);


};


#endif