
#ifndef _XMLFILEMANAGER_
#define _XMLFILEMANAGER_


#include "AppLogic.h"
#include "XmlMsgInterface.h"


////////////////////////////////////////////////////////////////////////////////////////
//
//
// CXmlFileManager class
//
//
/////////////////////////////////////////////////////////////////////////////////////////
// Stephane Rodriguez, 1st july 2002
//
//

class CXmlFileManager : public IXmlMsgInterface
{

	// Members
	//
protected:
	char	m_szURL[SAXML_URLMAXLENGTH];
	int		m_cnt;
	BOOL	m_bSkipNow;
	BOOL	m_bPromptMode; // if TRUE, optimization msg box is shown
	void*	m_cpParent; // used to send message (SetParamInt, ...)

	static long m_nLastError;

	// Constructor
	//
public:
	CXmlFileManager();
	virtual ~CXmlFileManager();

	// Methods
	//
	static  CXmlFileManager *chooseResourceManager(char *szURL);
	void	setPromptMode(BOOL bPrompt=TRUE);
	void	peekMessage();
	void	skipNow(BOOL bSkipNow=TRUE);
	BOOL	shouldSkipNow();

	virtual BOOL SetParamInt(int nParamID, int nValue);
	virtual BOOL SetParamString(int nParamID, char *szValue);
	virtual BOOL SetParamPtr(int nParamID, void *pValue);

	virtual int  GetParamInt(int nParamID);
	virtual char *GetParamString(int nParamID);
	virtual void *GetParamPtr(int nParamID);
 
	virtual void SetMsgHandler(void *pHandler);

	virtual BOOL Open(char *szURL) = 0; // VIRTUAL PURE
	virtual BOOL Read(char *szInputBuffer, int nMaxLength, int &nActualLength) = 0; // VIRTUAL PURE
	virtual BOOL Close() = 0; // VIRTUAL PURE

	virtual void  setParent(void *pParent);
	virtual void* getParent();


	static void setLastError(long nError);
	static long getLastError();


	// Prohibited use
	//
private:
	CXmlFileManager& operator=(const CXmlFileManager &obj);

};


class CXmlLocalFileManager : public CXmlFileManager
{
	// Members
	//
protected:
	FILE *fp; // file specific variable

	// Constructor
	//
public:
	CXmlLocalFileManager();
	virtual ~CXmlLocalFileManager();

	// Methods
	//
	virtual BOOL Open(char *szURL); // the URL is actually a fully qualified local filepath
	virtual BOOL Read(char *szInputBuffer, int nMaxLength, int &nActualLength);
	virtual BOOL Close();


	// Prohibited use
	//
private:
	CXmlLocalFileManager& operator=(const CXmlLocalFileManager &obj);
};


class CXmlInternetFileManager : public CXmlFileManager, public IBindStatusCallback
{
	// Members
	//
protected:
	long m_cRef; // object life cycle

	char		m_szProxy[MAX_PATH];
	int			m_nProxyPort;

	// IBindStatusCallback control
	IBinding*   m_pBinding;       // ptr to current IBinding
	IStream*    m_pStream;        // ptr to data stream
	
	HWND        m_hwndStatus;     // status line window
	FILE		*m_tmpFile;		  // temporary buffer

	BOOL		m_bStop_EvrythingIsReceived;
	DWORD       m_dwLoadedSoFar;  // num bytes transferred so far
	DWORD		m_dwOffsetReadByClient; // begin bytes in stream where client should collect data


	// Constructor
	//
public:
	CXmlInternetFileManager();
	virtual ~CXmlInternetFileManager();

	// Methods
	//
	virtual BOOL Open(char *szURL);
	virtual BOOL Read(char *szInputBuffer, int nMaxLength, int &nActualLength);
	virtual BOOL Close();


public:

	// IUnknown implementation
	virtual HRESULT __stdcall	QueryInterface(const IID& iid, void** ppv) ;			
	virtual ULONG __stdcall		AddRef() ;
	virtual ULONG __stdcall		Release() ;

	// NOTE: Not all IBindStatusCallback functions are shown
   //
        virtual HRESULT STDMETHODCALLTYPE OnStartBinding( 
            /* [in] */ DWORD dwReserved,
            /* [in] */ IBinding __RPC_FAR *pib);
        
        virtual HRESULT STDMETHODCALLTYPE GetPriority( 
            /* [out] */ LONG __RPC_FAR *pnPriority);
        
        virtual HRESULT STDMETHODCALLTYPE OnLowResource( 
            /* [in] */ DWORD reserved);
        
        virtual HRESULT STDMETHODCALLTYPE OnProgress( 
            /* [in] */ ULONG ulProgress,
            /* [in] */ ULONG ulProgressMax,
            /* [in] */ ULONG ulStatusCode,
            /* [in] */ LPCWSTR szStatusText);
        
        virtual HRESULT STDMETHODCALLTYPE OnStopBinding( 
            /* [in] */ HRESULT hresult,
            /* [unique][in] */ LPCWSTR szError);
        
        virtual /* [local] */ HRESULT STDMETHODCALLTYPE GetBindInfo( 
            /* [out] */ DWORD __RPC_FAR *grfBINDF,
            /* [unique][out][in] */ BINDINFO __RPC_FAR *pbindinfo);
        
        virtual /* [local] */ HRESULT STDMETHODCALLTYPE OnDataAvailable( 
            /* [in] */ DWORD grfBSCF,
            /* [in] */ DWORD dwSize,
            /* [in] */ FORMATETC __RPC_FAR *pformatetc,
            /* [in] */ STGMEDIUM __RPC_FAR *pstgmed);
        
        virtual HRESULT STDMETHODCALLTYPE OnObjectAvailable( 
            /* [in] */ REFIID riid,
            /* [iid_is][in] */ IUnknown __RPC_FAR *punk);

/*
   STDMETHODIMP OnStartBinding(IBinding* pbinding);
   STDMETHODIMP OnProgress(ULONG ulProgress, ULONG ulProgressMax, ULONG ulStatusCode, LPCWSTR pwzStatusText);
   STDMETHODIMP OnDataAvailable(DWORD grfBSCF, DWORD dwSize, 
                                FORMATETC *pFmtetc, REFIID riid, 
                                IUnknown* punk);
   STDMETHODIMP OnStopBinding(HRESULT hrResult);
*/

	// Prohibited use
	//
private:
	CXmlInternetFileManager& operator=(const CXmlInternetFileManager &obj);
};



class CXmlSecuredFileManager : public CXmlFileManager
{
	// Members
	//
protected:

	// Constructor
	//
public:
	CXmlSecuredFileManager();
	virtual ~CXmlSecuredFileManager();

	// Methods
	//
	virtual BOOL SetParamInt(int nParamID, int nValue);
	virtual BOOL SetParamString(int nParamID, char *szValue);
	virtual BOOL Open(char *szURL);
	virtual BOOL Read(char *szInputBuffer, int nMaxLength, int &nActualLength);
	virtual BOOL Close();

	// Prohibited use
	//
private:
	CXmlSecuredFileManager& operator=(const CXmlSecuredFileManager &obj);
};

#endif
